""" Measures of association between a Quantitative feature and binary target.
"""

from math import sqrt
from typing import Any

from numpy import nan
from pandas import DataFrame, Series
from scipy.spatial.distance import correlation
from scipy.stats import kruskal, pearsonr, spearmanr
from statsmodels.formula.api import ols


def kruskal_measure(
    x: Series,
    y: Series,
    thresh_kruskal: float = 0,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Kruskal-Wallis' test statistic between ``x`` for each value taken by ``y``.

    Parameters
    ----------
    x : Series
        Quantitative feature
    y : Series
        Qualitative target feature
    thresh_kruskal : float, optional
        Minimum Kruskal-Wallis association, by default ``0``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether ``x`` is sufficiently associated to ``y`` and Kruskal-Wallis' test statistic
    """
    nans = x.isnull()  # ckecking for nans

    # getting y values
    y_values = y.unique()

    # computation of Kruskal-Wallis statistic
    kw = kruskal(*tuple(x[(~nans) & (y == y_value)] for y_value in y_values))

    # updating association
    active, measurement = False, {"kruskal_measure": nan}
    if kw:
        measurement = {"kruskal_measure": kw[0]}

        # Excluding features not associated enough
        active = kw[0] < thresh_kruskal

    return active, measurement


def R_measure(
    x: Series,
    y: Series,
    thresh_R: float = 0,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Square root of the coefficient of determination of linear regression model of ``x`` by ``y``.

    Parameters
    ----------
    x : Series
        Quantitative feature
    y : Series
        Binary target feature
    thresh_R : float, optional
        Minimum R association, by default ``0``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether ``x`` is sufficiently associated to ``y`` and the square root of the determination coefficient
    """
    nans = x.isnull()  # ckecking for nans

    # grouping feature and target
    ols_df = DataFrame({"feature": x[~nans], "target": y[~nans]})

    # fitting regression of feature by target
    regression = ols("feature~C(target)", ols_df).fit()

    # updating association
    active, measurement = False, {"R_measure": nan}
    if regression.rsquared and regression.rsquared >= 0:
        r_measure = sqrt(regression.rsquared)
        measurement = {"R_measure": r_measure}

        # Excluding features not associated enough
        active = r_measure < thresh_R

    return active, measurement


def pearson_measure(
    x: Series,
    y: Series,
    thresh_pearson: float = 0,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Pearson's linear correlation coefficient between ``x`` and ``y``.

    Parameters
    ----------
    x : Series
        Quantitative feature
    y : Series
        Quantitative target feature
    thresh_pearson : float, optional
        Minimum r association, by default ``0``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether ``x`` is sufficiently associated to ``y`` and Pearson's r
    """
    nans = x.isnull()  # ckecking for nans

    # computing spearman's r
    r = pearsonr(x[~nans], y[~nans])

    # updating association
    active, measurement = False, {"pearson_measure": nan}
    if r:
        measurement = {"pearson_measure": r}

        # Excluding features not associated enough
        active = r < thresh_pearson

    return active, measurement


def spearman_measure(
    x: Series,
    y: Series,
    thresh_spearman: float = 0,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Spearman's rank correlation coefficient between ``x`` and ``y``.

    Parameters
    ----------
    x : Series
        Quantitative feature
    y : Series
        Quantitative target feature
    thresh_spearman : float, optional
        Minimum rho association, by default ``0``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether ``x`` is sufficiently associated to ``y`` and Spearman's rho
    """
    nans = x.isnull()  # ckecking for nans

    # computing spearman's r
    rho = spearmanr(x[~nans], y[~nans])

    # updating association
    active, measurement = False, {"spearman_measure": nan}
    if rho:
        measurement = {"spearman_measure": rho}

        # Excluding features not associated enough
        active = rho < thresh_spearman

    return active, measurement


def distance_measure(
    x: Series,
    y: Series,
    thresh_distance: float = 0,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Distance correlation between ``x`` and ``y``.

    Parameters
    ----------
    x : Series
        Quantitative feature
    y : Series
        Quantitative target feature
    thresh_distance : float, optional
        Minimum distance association, by default ``0``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether ``x`` is sufficiently associated to ``y`` and Distance Correlation
    """
    nans = x.isnull()  # ckecking for nans

    # computing spearman's r
    d_corr = correlation(x[~nans], y[~nans])

    # updating association
    active, measurement = False, {"distance_measure": nan}
    if d_corr:
        measurement = {"distance_measure": d_corr}

        # Excluding features not associated enough
        active = d_corr < thresh_distance

    return active, measurement


def zscore_measure(
    x: Series,
    y: Series = None,
    thresh_zscore: float = 1.0,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Computes outliers percentage based on the z-score

    Parameters
    ----------
    x : Series
        Quantitative feature
    y : Series, optional
        Any target feature, by default ``None``
    thresh_zscore : float, optional
        Maximum percentage of Outliers in a feature, by default ``1.0``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether or not there are too many outliers and the outlier measurement
    """
    mean = x.mean()  # mean of the feature
    std = x.std()  # standard deviation of the feature
    zscore = (x - mean) / std  # zscore per observation

    # checking for outliers
    outliers = abs(zscore) > 3
    pct_zscore = outliers.mean()

    # updating association
    measurement = {
        "pct_zscore": pct_zscore,
        "min": x.min(),
        "max": x.max(),
        "mean": mean,
        "std": std,
    }

    # Excluding feature with too frequent modes
    active = pct_zscore < thresh_zscore

    return active, measurement


def iqr_measure(
    x: Series,
    y: Series = None,
    thresh_iqr: float = 1.0,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Computes outliers percentage based on the interquartile range

    Parameters
    ----------
    x : Series
        Quantitative feature
    y : Series, optional
        Any target feature, by default ``None``
    thresh_iqr : float, optional
        Maximum percentage of Outliers in a feature, by default ``1.0``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether or not there are too many outliers and the outlier measurement
    """
    q3 = x.quantile(0.75)  # 3rd quartile
    q1 = x.quantile(0.25)  # 1st quartile
    iqr = q3 - q1  # inter quartile range
    iqr_bounds = q1 - 1.5 * iqr, q3 + 1.5 * iqr  # bounds of the iqr range

    # checking for outliers
    outliers = ~x.between(*iqr_bounds)
    pct_iqr = outliers.mean()

    # updating association
    measurement = {"pct_iqr": pct_iqr, "q1": q1, "median": x.median(), "q3": q3}

    # Excluding feature with too frequent modes
    active = pct_iqr < thresh_iqr

    return active, measurement



"""Tools to select the best Quantitative and Qualitative features."""

from random import shuffle
from typing import Any, Callable, Union
from warnings import warn

from numpy import unique
from pandas import DataFrame, Series

from .filters import thresh_filter
from .measures import dtype_measure, make_measure, mode_measure, nans_measure

# trying to import extra dependencies
try:
    from IPython.display import display_html
except ImportError:
    _has_idisplay = False
else:
    _has_idisplay = True


class BaseSelector:
    """A pipeline of measures to perform a feature pre-selection that maximizes association
    with a binary target.

    * Best features are the n_best of each measure
    * Get your best features with ``Selector.select()``!
    """

    def __init__(
        self,
        n_best: int,
        quantitative_features: list[str] = None,
        qualitative_features: list[str] = None,
        *,
        measures: Union[list[Callable], dict[str, list[Callable]]] = None,
        filters: Union[list[Callable], dict[str, list[Callable]]] = None,
        colsample: float = 1.0,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        # measures : list[Callable], optional
        #     List of association measures to be used, by default ``None``.
        #     Ranks features based on last provided measure of the list.
        #     See :ref:`Measures`.
        #     Implemented measures are:

        #     * [Quantitative Features] For association evaluation: ``kruskal_measure`` (default), ``R_measure``
        #     * [Quantitative Features] For outlier detection: ``zscore_measure``, ``iqr_measure``
        #     * [Qualitative Features] For association evaluation: ``chi2_measure``, ``cramerv_measure``, ``tschuprowt_measure`` (default)

        # filters : list[Callable], optional
        #     List of filters to be used, by default ``None``.
        #     See :ref:`Filters`.
        #     Implemented filters are:

        #     * [Quantitative Features] For linear correlation: ``spearman_filter`` (default), ``pearson_filter``
        #     * [Qualitative Features] For correlation: ``cramerv_filter``, ``tschuprowt_filter`` (default)

        """
        Parameters
        ----------
        n_best : int
            Number of features to select.

            * Best features are the ``n_best`` of each provided data types (set in ``quantitative_features`` and/or ``qualitative_features``)
            * Best features are the ``n_best`` for each provided measures (set in ``quantitative_measures`` and/or ``qualitative_measures``)

        quantitative_features : list[str]
            List of column names of quantitative features to chose from, by default ``None``.
            Must be set if ``qualitative_features=None``.

        qualitative_features : list[str]
            List of column names of qualitative features to chose from, by default ``None``.
            Must be set if ``quantitative_features=None``.

        quantitative_measures : list[Callable], optional
            List of association measures to be used for ``quantitative_features``.
            :ref:`Implemented measures <Measures>` are:

            * For association evaluation: :ref:`Kruskal-Wallis' H <kruskal>` (default), :ref:`R`
            * For outlier detection: :ref:`Standard score <zscore>`, :ref:`Interquartile range <iqr>`

        qualitative_measures : list[Callable], optional
            List of association measures to be used for ``qualitative_features``.
            :ref:`Implemented measures <Measures>` are:

            * For association evaluation: :ref:`Pearson's chi² <chi2>`, :ref:`cramerv`, :ref:`tschuprowt` (default)

        quantitative_filters : list[Callable], optional
            List of filters to be used for ``quantitative_features``.
            :ref:`Implemented filters <Filters>` are:

            * For linear correlation: :ref:`pearson_filter`, :ref:`Spearman's rho <spearman_filter>` (default)

        qualitative_filters : list[Callable], optional
            List of filters to be used for ``qualitative_features``.
            :ref:`Implemented filters <Filters>` are:

            * For correlation: :ref:`cramerv_filter`, :ref:`tschuprowt_filter` (default)

        colsample : float, optional
            Size of sampled list of features for sped up computation, between ``0`` and ``1``, by default ``1.0``
            By default, all features are used.

            For ``colsample=0.5``, Selector will search for the best features in
            ``features[:len(features)//2]`` and then in ``features[len(features)//2:]``.

            **Tip:** for better performance, should be set such as ``len(features)//2 < 200``.

        verbose : bool, optional
            * ``True``, without IPython installed: prints raw feature selection steps for X, by default ``False``
            * ``True``, with IPython installed: adds HTML tables to the output.

            **Tip**: IPython displaying can be turned off by setting ``pretty_print=False``.

        **kwargs
            Allows one to set thresholds for provided ``quantitative_measures``/``qualitative_measures`` and ``quantitative_filters``/``qualitative_filters`` (see :ref:`Measures` and :ref:`Filters`) passed as keyword arguments.

        Examples
        --------
        See `Selectors examples <https://autocarver.readthedocs.io/en/latest/index.html>`_
        """
        # settinp up list of features
        if quantitative_features is None:
            quantitative_features = []
        self.quantitative_features = list(set(quantitative_features))
        if qualitative_features is None:
            qualitative_features = []
        self.qualitative_features = list(set(qualitative_features))
        assert len(quantitative_features) > 0 or len(qualitative_features) > 0, (
            "No feature passed as input. Pleased provided column names to Carver by setting "
            "qualitative_features or quantitative_features."
        )
        self.features = list(set(self.qualitative_features + self.quantitative_features))
        self.input_dtypes = {"float": self.quantitative_features, "str": self.qualitative_features}

        # number of features selected
        self.n_best = n_best
        assert (
            0 < int(self.n_best) <= len(self.features) + 1
        ), "Must set 0 < n_best <= len(features)"

        # feature sample size per iteration
        self.colsample = colsample

        # adding default measures
        self.measures = {
            dtype: [dtype_measure, nans_measure, mode_measure] + requested_measures[:]
            for dtype, requested_measures in measures.items()
        }

        # adding default filter
        self.filters = {
            dtype: [thresh_filter] + requested_filters[:]
            for dtype, requested_filters in filters.items()
        }

        # names of measures to sort by
        self.measure_names = {
            dtype: [measure.__name__ for measure in requested_measures[::-1]]
            for dtype, requested_measures in self.measures.items()
        }

        # wether or not to print tables
        self.verbose = bool(max(verbose, kwargs.get("pretty_print", False)))
        self.pretty_print = False
        if self.verbose and kwargs.get("pretty_print", True):
            if _has_idisplay:  # checking for installed dependencies
                self.pretty_print = True
            else:
                warn(
                    "Package not found: IPython. Defaulting to raw verbose. "
                    "Install extra dependencies with pip install autocarver[jupyter]",
                    UserWarning,
                )

        # keyword arguments
        self.kwargs = kwargs

    def _select_features(
        self,
        X: DataFrame,
        y: Series,
        features: list[str],
        n_best: int,
        dtype: str,
    ) -> list[str]:
        """Selects the n_best features amongst the specified ones"""
        if self.verbose:  # verbose if requested
            data_type = "quantitative"
            if dtype != "float":
                data_type = "qualitative"
            print(f"------\n[Selector] Selecting from {data_type} features: {str(features)}\n---")

        # Computes association between X and y
        initial_associations = apply_measures(
            X, y, measures=self.measures[dtype], features=features, **self.kwargs
        )

        # sorting statistics
        measure_names = evaluated_measure_names(initial_associations, self.measure_names[dtype])
        initial_associations = initial_associations.sort_values(measure_names, ascending=False)

        # displaying non-filterd associations
        self._print_associations(association=initial_associations)

        # applying filtering for each measure
        all_best_features: dict[str, Any] = {}
        for measure_name in measure_names:
            # sorting association for each measure
            associations = initial_associations.sort_values(measure_name, ascending=False)

            # filtering for each measure, as each measure ranks the features differently
            filtered_association = apply_filters(
                X, associations, filters=self.filters[dtype], **self.kwargs
            )

            # selected features for the measure
            selected_features = [
                feature
                for feature in initial_associations.index
                if (len(filtered_association) > 0)
                and (feature in filtered_association.index[:n_best])
            ]

            # saving results
            all_best_features.update(
                {
                    measure_name: {
                        "selected": selected_features,
                        "association": filtered_association,
                    }
                }
            )

        # list of unique best_featues per measure_name
        best_features = [
            # ordering according target association
            feature
            for feature in initial_associations.index
            # checking that feature has been selected by a measure
            if any(feature in all_best_features[measure]["selected"] for measure in measure_names)
        ]

        # displaying association measure
        self._print_associations(
            association=initial_associations.reindex(best_features),
            message=", filtered for inter-feature assocation",
        )

        return best_features

    def select(self, X: DataFrame, y: Series) -> list[str]:
        """Selects the ``n_best`` features of the DataFrame, by association with the target

        Parameters
        ----------
        X : DataFrame
            Dataset to determine optimal features.
            Needs to have columns has specified in ``features`` attribute.

        y : Series
            Target with wich the association is evaluated.

        Returns
        -------
        list[str]
            List of selected features
        """
        # iterating over each type of feature
        all_best_features = []
        for dtype in unique(list(self.input_dtypes.keys())):
            # getting features of the specific data type
            features = self.input_dtypes[dtype][:]

            # splitting features in chunks
            if self.colsample < 1:
                # shuffling features to get random samples of features
                shuffle(features)

                # number of features per sample
                chunks = int(len(self.features) // (1 / self.colsample))

                # splitting feature list in samples
                feature_samples = [
                    features[chunks * i : chunks * (i + 1)]
                    for i in range(int(1 / self.colsample) - 1)
                ]

                # adding last sample with all remaining features
                feature_samples += [features[chunks * (int(1 / self.colsample) - 1) :]]

                # iterating over each feature samples
                best_features = []
                for feature_sample in feature_samples:
                    # fitting association on features
                    best_features += self._select_features(
                        X, y, feature_sample, int(self.n_best // 2), dtype
                    )

            # splitting in chunks not requested
            else:
                best_features = features[:]

            # final selection with all best_features selected
            if any(best_features):
                best_features = self._select_features(X, y, best_features, self.n_best, dtype)
                all_best_features += best_features
                if self.verbose:  # verbose if requested
                    data_type = "quantitative"
                    if dtype != "float":
                        data_type = "qualitative"
                    print(f"\n - [Selector] Selected {data_type} features: {str(best_features)}")
                    print("------\n")

        return all_best_features

    def _print_associations(self, association: DataFrame, message: str = "") -> None:
        """EDA of fitted associations

        Parameters
        ----------
        association : DataFrame
            _description_
        pretty_print : bool, optional
            _description_, by default False
        """
        # checking for verbose
        if self.verbose:
            print(f"\n - [Selector] Association between X and y{message}")

            # printing raw DataFrame
            if not self.pretty_print:
                print(association)

            # adding colors and displaying DataFrame as html
            else:
                # finding columns with indicators to colorize
                subset = [
                    column
                    for column in association.columns
                    # checking for an association indicator
                    if any(indic in column for indic in ["pct_", "_measure", "_filter"])
                ]
                # getting prettier association table
                nicer_association = association.style.background_gradient(
                    cmap="coolwarm", subset=subset
                )
                nicer_association = nicer_association.set_table_attributes("style='display:inline'")

                # displaying html of colored DataFrame
                display_html(nicer_association._repr_html_(), raw=True)


def feature_association(x: Series, y: Series, measures: list[Callable], **kwargs) -> dict[str, Any]:
    """Measures association between x and y

    Parameters
    ----------
    x : Series
        Sample of a feature.
    y : Series
        Binary target feature with wich the association is evaluated.

    Returns
    -------
    dict[str, Any]
        Association metrics' values
    """
    # measures keep going only if previous basic tests are passed
    passed = True
    association = {}

    # iterating over each measure
    for measure in measures:
        passed, association = make_measure(
            measure, passed, association, x, y, **kwargs, **association
        )

    return association


def apply_measures(
    X: DataFrame, y: Series, measures: list[Callable], features: list[str], **kwargs
) -> DataFrame:
    """Measures association between columns of X and y

    Parameters
    ----------
    X : DataFrame
        _description_
    y : Series
        _description_
    features : list[str]
        _description_
    measure_names : list[str]
        _description_

    Returns
    -------
    DataFrame
        _description_
    """
    # applying association measure to each column
    associations = (
        X[features]
        .apply(feature_association, y=y, measures=measures, **kwargs, result_type="expand", axis=0)
        .T
    )

    return associations


def evaluated_measure_names(associations: DataFrame, measure_names: list[str]) -> list[str]:
    """_summary_

    Parameters
    ----------
    associations : DataFrame
        _description_
    measure_names : list[str]
        _description_

    Returns
    -------
    list[str]
        _description_
    """
    # Getting evaluated measures (filtering out non-measures: pct_zscore, pct_iqr...)
    sort_by = [
        measure_name
        for measure_name in measure_names
        if measure_name in associations and "_measure" in measure_name
    ]

    return sort_by


def apply_filters(
    X: DataFrame, associations: DataFrame, filters: list[Callable], **kwargs
) -> DataFrame:
    """Filters out too correlated features (least relevant first)

    Parameters
    ----------
    X : DataFrame
        _description_
    associations : DataFrame
        _description_
    filters : list[Callable]
        _description_
    measure_name : str
        _description_

    Returns
    -------
    DataFrame
        _description_
    """
    # applying successive filters
    filtered_associations = associations.copy()
    for filtering in filters:
        filtered_associations = filtering(X, filtered_associations, **kwargs)

    return filtered_associations



"""Tools to build simple buckets out of Qualitative features
for a binary classification model.
"""

from typing import Any, Union
from warnings import warn

from numpy import argmin, nan, select
from pandas import DataFrame, Series, isna, notna, unique

from .base_discretizers import (
    BaseDiscretizer,
    convert_to_labels,
    convert_to_values,
    extend_docstring,
    nan_unique,
    target_rate,
    value_counts,
)
from .grouped_list import GroupedList
from .type_discretizers import StringDiscretizer


class CategoricalDiscretizer(BaseDiscretizer):
    """Automatic discretization of categorical features, building simple groups frequent enough.

    Groups a qualitative features' values less frequent than ``min_freq`` into a ``str_default`` string.

    NaNs are left untouched.

    Only use for qualitative non-ordinal features.
    """

    @extend_docstring(BaseDiscretizer.__init__)
    def __init__(
        self,
        qualitative_features: list[str],
        min_freq: float,
        *,
        values_orders: dict[str, GroupedList] = None,
        copy: bool = False,
        verbose: bool = False,
        str_default: str = "__OTHER__",
        str_nan: str = "__NAN__",
    ) -> None:
        """
        Parameters
        ----------
        qualitative_features : list[str]
            List of column names of qualitative features (non-ordinal) to be discretized

        min_freq : float
            Minimum frequency per grouped modalities.

            * Features whose most frequent modality is less frequent than ``min_freq`` will not be discretized.
            * Sets the number of quantiles in which to discretize the continuous features.
            * Sets the minimum frequency of a quantitative feature's modality.

            **Tip**: should be set between ``0.02`` (slower, preciser, less robust) and ``0.05`` (faster, more robust)

        str_default : str, optional
            String representation for default qualitative values, i.e. values less frequent than ``min_freq``, by default ``"__OTHER__"``
        """
        # Initiating BaseDiscretizer
        super().__init__(
            features=qualitative_features,
            values_orders=values_orders,
            input_dtypes="str",
            output_dtype="str",
            str_nan=str_nan,
            str_default=str_default,
            copy=copy,
            verbose=verbose,
        )

        self.min_freq = min_freq

    def _prepare_data(self, X: DataFrame, y: Series) -> DataFrame:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in ``CategoricalDiscretizer.features``.

        y : Series
            Binary target feature with wich the association is maximized.

        Returns
        -------
        DataFrame
            A formatted copy of X
        """
        # checking for binary target
        x_copy = super()._prepare_data(X, y)

        # checks and initilizes values_orders
        for feature in self.qualitative_features:
            # initiating features missing from values_orders
            if feature not in self.values_orders:
                self.values_orders.update({feature: GroupedList(nan_unique(x_copy[feature]))})

        # checking that all unique values in X are in values_orders
        self._check_new_values(x_copy, features=self.qualitative_features)

        # adding NANS
        for feature in self.qualitative_features:
            order = self.values_orders[feature]
            if any(x_copy[feature].isna()):
                if self.str_nan not in order.values():
                    order.append(self.str_nan)
                    self.values_orders.update({feature: order})

        # filling up NaNs
        x_copy[self.qualitative_features] = x_copy[self.qualitative_features].fillna(self.str_nan)

        return x_copy

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series) -> None:
        # copying dataframe and checking data before bucketization
        x_copy = self._prepare_data(X, y)

        if self.verbose:  # verbose if requested
            print(f" - [CategoricalDiscretizer] Fit {str(self.qualitative_features)}")

        # computing frequencies of each modality
        frequencies = x_copy[self.qualitative_features].apply(value_counts, normalize=True, axis=0)

        for feature in self.qualitative_features:
            # sorting orders based on target rates
            order = self.values_orders[feature]

            # checking for rare values
            values_to_group = [
                val
                for val, freq in frequencies[feature].items()
                if freq < self.min_freq and val != self.str_nan
            ]

            # adding values that are completly missing (no frequency in X)
            values_to_group += [value for value in order if value not in frequencies[feature]]

            # grouping values to str_default if any
            if any(values_to_group):
                # adding default value to the order
                order.append(self.str_default)

                # grouping rare values in default value
                order.group_list(values_to_group, self.str_default)
                x_copy.loc[x_copy[feature].isin(values_to_group), feature] = self.str_default

                # updating values_orders
                self.values_orders.update({feature: order})

        # computing target rate per modality for ordering
        target_rates = x_copy[self.qualitative_features].apply(
            target_rate, y=y, ascending=True, axis=0
        )

        # sorting orders based on target rates
        for feature in self.qualitative_features:
            order = self.values_orders[feature]

            # new ordering according to target rate
            new_order = list(target_rates[feature])

            # checking for default but no value observed, enable to group this modality, raising error
            assert (self.str_default in order and self.str_default in new_order) or (
                self.str_default not in order and self.str_default not in new_order
            ), f"Some values from values_orders['{feature}'] are never observed. Can not fit a distribution without any observation. Please remove following values {str([value for value in order.content[self.str_default] if value != self.str_default])} from values_orders['{feature}']."

            # leaving NaNs at the end of the list
            if self.str_nan in new_order:
                new_order.remove(self.str_nan)
                new_order += [self.str_nan]

            # sorting order updating values_orders
            self.values_orders.update({feature: order.sort_by(new_order)})

        # discretizing features based on each feature's values_order
        super().fit(X, y)

        return self


class OrdinalDiscretizer(BaseDiscretizer):
    """Automatic discretization of ordinal features, grouping less frequent modalities with the
    closest modlity in target rate or by frequency.

    NaNs are left untouched.

    Only use for qualitative ordinal features.

    Fisrt fits :ref:`StringDiscretizer` if neccesary.
    """

    @extend_docstring(BaseDiscretizer.__init__)
    def __init__(
        self,
        ordinal_features: list[str],
        min_freq: float,
        values_orders: dict[str, GroupedList],
        *,
        input_dtypes: Union[str, dict[str, str]] = "str",
        copy: bool = False,
        verbose: bool = False,
        str_nan: str = "__NAN__",
    ):
        """
        Parameters
        ----------
        ordinal_features : list[str]
            List of column names of ordinal features to be discretized. For those features a list of values has to be provided in the ``values_orders`` dict.

        min_freq : float
            Minimum frequency per grouped modalities.

            * Features whose most frequent modality is less frequent than ``min_freq`` will not be discretized.
            * Sets the number of quantiles in which to discretize the continuous features.
            * Sets the minimum frequency of a quantitative feature's modality.

            **Tip**: should be set between ``0.02`` (slower, preciser, less robust) and ``0.05`` (faster, more robust)

        input_dtypes : Union[str, dict[str, str]], optional
            Input data type, converted to a dict of the provided type for each feature, by default ``"str"``

            * ``"str"``, features are considered as qualitative.
            * ``'float"``, features are considered as quantitative.
        """
        # Initiating BaseDiscretizer
        super().__init__(
            features=ordinal_features,
            values_orders=values_orders,
            input_dtypes=input_dtypes,
            output_dtype="str",
            str_nan=str_nan,
            copy=copy,
            verbose=verbose,
        )

        # checking for missong orders
        no_order_provided = [
            feature for feature in self.features if feature not in self.values_orders
        ]
        assert len(no_order_provided) == 0, (
            " - [OrdinalDiscretizer] No ordering was provided for following features:"
            f" {str(no_order_provided)}. Please make sure you defined ``values_orders`` correctly."
        )

        # class specific attributes
        self.min_freq = min_freq

    def _prepare_data(self, X: DataFrame, y: Series) -> DataFrame:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in ``OrdinalDiscretizer.features``.

        y : Series
            Binary target feature.

        Returns
        -------
        DataFrame
            A formatted copy of X
        """
        # checking for binary target and copying X
        x_copy = super()._prepare_data(X, y)
        # adding NANS
        for feature in self.features:
            if any(x_copy[feature].isna()):
                values = self.values_orders[feature]
                if not values.contains(self.str_nan):
                    values.append(self.str_nan)
                    self.values_orders.update({feature: values})

        # removing NaNs if any already imputed -> grouping only non-nan values
        if self.str_nan:
            x_copy = x_copy.replace(self.str_nan, nan)

        return x_copy

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series) -> None:
        if self.verbose:  # verbose if requested
            print(f" - [OrdinalDiscretizer] Fit {str(self.features)}")

        # checking values orders
        x_copy = self._prepare_data(X, y)

        # converting potential quantiles into there labels
        known_orders = convert_to_labels(
            features=self.features,
            quantitative_features=self.quantitative_features,
            values_orders=self.values_orders,
            str_nan=self.str_nan,
            dropna=True,
        )

        # grouping rare modalities for each feature
        common_modalities = x_copy[self.features].apply(
            find_common_modalities,
            y=y,
            min_freq=self.min_freq,
            values_orders=known_orders,
            axis=0,
            result_type="reduce",
        )

        # converting potential labels into there respective values (quantiles)
        self.values_orders.update(
            convert_to_values(
                features=self.features,
                quantitative_features=self.quantitative_features,
                values_orders=self.values_orders,
                label_orders=common_modalities,
                str_nan=self.str_nan,
            )
        )

        # discretizing features based on each feature's values_order
        super().fit(x_copy, y)

        return self


class ChainedDiscretizer(BaseDiscretizer):
    """Automatic discretization of categorical features, joining rare modalities into higher
    level groups.

    For each provided :class:`GroupedList` from ``chained_orders`` attribute, values less frequent than
    ``min_freq`` are grouped in there respective group, as defined by :class:`GroupedList`.
    """

    @extend_docstring(BaseDiscretizer.__init__)
    def __init__(
        self,
        qualitative_features: list[str],
        min_freq: float,
        chained_orders: list[GroupedList],
        *,
        values_orders: dict[str, GroupedList] = None,
        unknown_handling: str = "raise",
        copy: bool = False,
        verbose: bool = False,
        str_nan: str = "__NAN__",
    ) -> None:
        """
        Parameters
        ----------
        qualitative_features : list[str]
            List of column names of qualitative features (non-ordinal) to be discretized

        chained_orders : list[GroupedList]
            A list of interlocked higher level groups for each modalities of each ordianl feature.
            Values of ``chained_orders[0]`` have to be grouped in ``chained_order[1]`` etc.

        min_freq : float
            Minimum frequency per grouped modalities.

            * Features whose most frequent modality is less frequent than ``min_freq`` will not be discretized.
            * Sets the number of quantiles in which to discretize the continuous features.
            * Sets the minimum frequency of a quantitative feature's modality.

            **Tip**: should be set between ``0.02`` (slower, preciser, less robust) and ``0.05`` (faster, more robust)

        unknown_handling : str, optional
            Whether or not to remove unknown values, by default ``'raise'``.

            * ``'raise'``, unknown values raise an ``AssertionError``.
            * ``'drop'``, unknown values are grouped with ``str_nan``.
        """
        # Initiating BaseDiscretizer
        super().__init__(
            features=qualitative_features,
            values_orders=values_orders,
            input_dtypes="str",
            output_dtype="str",
            str_nan=str_nan,
            dropna=False,
            copy=copy,
            verbose=verbose,
        )

        # class specific attributes
        self.min_freq = min_freq
        self.chained_orders = [GroupedList(values) for values in chained_orders]
        assert unknown_handling in [
            "drop",
            "raise",
        ], " - [ChainedDiscretizer] Wrong value for attribute unknown_handling. Choose from ['drop', 'raise']."
        self.unknown_handling = unknown_handling

        # known_values: all ordered values describe in each level of the chained_orders
        # starting off with first level
        known_values = self.chained_orders[0].values()
        # adding each level
        for n, next_level in enumerate(self.chained_orders[1:]):
            # iterating over each group of the next level
            for next_group, next_values in next_level.content.items():
                # looking for known and unknwon values in next_level

                # checking for unknown values
                next_unknown = [
                    value
                    for value in next_values
                    if value not in known_values and value != next_group
                ]
                assert (
                    len(next_unknown) == 0
                ), f" - [ChainedDiscretizer] Values {str(next_unknown)}, provided in chained_orders[{n+1}] are missing from chained_orders[{n}]. Please make sure values are kept trhough each level."

                # checking for known values
                next_known = [
                    value for value in next_values if value in known_values and value != next_group
                ]
                assert (
                    len(next_known) > 0
                ), f" - [ChainedDiscretizer] For key '{next_group}', the provided chained_orders[{n+1}] has no values from chained_orders[:{n+1}]. Please provide some values existing values."

                # index of the highest ranked known value of the next_group
                highest_index = known_values.index(next_known[-1])

                # adding next_group to the order at the right place in the amongst known_values
                known_values = (
                    known_values[: highest_index + 1]
                    + [next_group]
                    + known_values[highest_index + 1 :]
                )

        # saving resulting known values
        self.known_values = known_values

        # adding known_values to each feature's order
        for feature in self.features:
            # checking for already known values of the feature
            if feature in self.values_orders:
                order = self.values_orders[feature]
            # no known values for the feature
            else:
                order = GroupedList([])

            # checking that all values from the order are in known_values
            for value in order:
                assert (
                    value in self.known_values
                ), f" - [ChainedDiscretizer] Value {value} from feature {feature} provided in values_orders is missing from levels of chained_orders. Add value to a level of chained_orders or adapt values_orders."

            # adding known values if missing from the order
            for value in self.known_values:
                if value not in order.values():
                    order.append(value)
            order = order.sort_by(self.known_values)
            self.values_orders.update({feature: order})

        self.verbose = verbose

    def _prepare_data(self, X: DataFrame, y: Series = None) -> DataFrame:
        """Validates format and content of X and y. Converts non-string columns into strings.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in ``ChainedDiscretizer.features``.

        y : Series
            Binary target feature, not used, by default None.

        Returns
        -------
        DataFrame
            A formatted copy of X
        """
        # copying dataframe
        x_copy = X.copy()

        # checking for binary target and previous fit
        x_copy = super()._prepare_data(x_copy, y)

        # checking for ids (unique value per row)
        max_frequencies = x_copy[self.features].apply(
            lambda u: u.value_counts(normalize=True, dropna=False).drop(nan, errors="ignore").max(),
            axis=0,
        )
        # for each feature, checking that at least one value is more frequent than min_freq
        all_features = self.features[:]
        for feature in all_features:
            if max_frequencies[feature] < self.min_freq:
                warn(
                    f" - [ChainedDiscretizer] For feature '{feature}', the largest modality"
                    f" has {max_frequencies[feature]:2.2%} observations which is lower than "
                    "min_freq={self.min_freq:2.2%}. This feature will not be Discretized. Consider"
                    " decreasing parameter min_freq or removing this feature.",
                    UserWarning,
                )
                super()._remove_feature(feature)

        # checking for columns containing floats or integers even with filled nans
        dtypes = x_copy[self.features].fillna(self.str_nan).applymap(type).apply(unique)
        not_object = dtypes.apply(lambda u: any(typ != str for typ in u))

        # non qualitative features detected
        if any(not_object):
            features_to_convert = list(not_object.index[not_object])
            if self.verbose:
                unexpected_dtypes = [
                    typ for dtyp in dtypes[not_object] for typ in dtyp if typ != str
                ]
                warn(
                    f" - [ChainedDiscretizer] Non-string features: {str(features_to_convert)}. Try"
                    "ing to convert them using type_discretizers.StringDiscretizer, otherwise conv"
                    f"ert them manually. Unexpected data types: {str(list(unexpected_dtypes))}.",
                    UserWarning,
                )

            # converting specified features into qualitative features
            stringer = StringDiscretizer(
                qualitative_features=features_to_convert, values_orders=self.values_orders
            )
            x_copy = stringer.fit_transform(x_copy, y)

            # updating values_orders accordingly
            self.values_orders.update(stringer.values_orders)

        # filling nans
        x_copy[self.features] = x_copy[self.features].fillna(self.str_nan)

        # adding nans and unknown values
        for feature in self.features:
            order = self.values_orders[feature]
            # checking for unknown values (missing from known_values)
            unknown_values = [
                value
                for value in x_copy[feature].unique()
                if value not in self.known_values and value != self.str_nan
            ]

            # converting unknown values to NaN
            if len(unknown_values) > 0:
                # raising an error
                if self.unknown_handling == "raise":
                    assert (
                        not len(unknown_values) > 0
                    ), f" - [ChainedDiscretizer] Order for feature '{feature}' needs to be provided for values: {str(unknown_values)}, otherwise set remove_unknown='drop' (policy unknown_handling='raise')"

                # dropping unknown value
                else:  # unknown_handling='drop'
                    # alerting user
                    print(
                        f" - [ChainedDiscretizer] Order for feature '{feature}' was not provided for values:  {str(unknown_values)}, these values will be converted to '{self.str_nan}' (policy unknown_handling='drop')"
                    )

                    # adding unknown to the order
                    for unknown_value in unknown_values:
                        order.append(unknown_value)
                        order.append(self.str_nan)
                        # grouping unknown value with str_nan
                        order.group(unknown_value, self.str_nan)

            # updating values_orders accordingly
            self.values_orders.update({feature: order})

        # adding up NAN for all features of values_orders for seamless integration
        # when GroupedList._tranform_qualitative is called nans are replaced by str_nan
        for feature, order in self.values_orders.items():
            # adding NaNs to the order if any
            if any(x_copy[feature].isna()) or any(x_copy[feature] == self.str_nan):
                if self.str_nan not in order:
                    order.append(self.str_nan)
                    self.values_orders.update({feature: order})

        # checking that all unique values in X are in values_orders
        self._check_new_values(x_copy, features=self.features)

        return x_copy

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series = None) -> None:
        # filling nans
        x_copy = self._prepare_data(X, y)

        if self.verbose:  # verbose if requested
            print(f" - [ChainedDiscretizer] Fit {str(self.features)}")

        # iterating over each feature
        for feature in self.features:
            # computing frequencies of each modality
            frequencies = x_copy[feature].value_counts(normalize=True)

            # iterating over each specified orders
            for level_order in self.chained_orders:
                values, frequencies = frequencies.index, frequencies.values

                # values that are frequent enough
                to_keep = list(values[frequencies >= self.min_freq]) + [
                    self.str_nan,
                ]

                # values from the order to group (not frequent enough or absent)
                values_to_group = [value for value in level_order.values() if value not in to_keep]

                # values to group into discarded values
                groups_value = [level_order.get_group(value) for value in values_to_group]

                # values of the feature to input (needed for next levels of the order)
                df_to_input = [x_copy[feature] == discarded for discarded in values_to_group]

                # inputing non frequent values
                x_copy[feature] = select(df_to_input, groups_value, default=x_copy[feature])

                # historizing in the feature's order
                for discarded, kept in zip(values_to_group, groups_value):
                    self.values_orders.get(feature).group(discarded, kept)

                # updating frequencies of each modality for the next ordering
                frequencies = x_copy[feature].value_counts(normalize=True)

        super().fit(X, y)

        return self


def find_common_modalities(
    df_feature: Series,
    y: Series,
    min_freq: float,
    values_orders: dict[str, GroupedList],
    len_df: int = None,
) -> dict[str, Any]:
    """Recursively finds the closest modality by target rate or by frequency.

    Parameters
    ----------
    df_feature : Series
        _description_
    y : Series
        _description_
    min_freq : float
        _description_
    values_orders : dict[str, GroupedList]
        _description_
    len_df : int, optional
        _description_, by default None

    Returns
    -------
    dict[str, Any]
        _description_
    """
    # getting feature's order
    order = values_orders[df_feature.name]

    # initialisation de la taille totale du dataframe
    if len_df is None:
        len_df = len(df_feature)

    # case 1: there are missing values
    if any(isna(df_feature)):
        return find_common_modalities(
            df_feature[notna(df_feature)],
            y[notna(df_feature)],
            min_freq,
            values_orders,
            len_df,
        )

    # case 2: no missing values
    # computing frequencies and target rate of each modality
    init_frequencies = df_feature.value_counts(dropna=False, normalize=False) / len_df
    init_target_rates = y.groupby(df_feature).sum().reindex(order)

    # per-modality/value frequencies, filling missing by 0
    frequencies = init_frequencies.reindex(order, fill_value=0)

    # case 2.1: there are underrepresented modalities/values
    while any(frequencies < min_freq) & (len(frequencies) > 1):
        # updating per-group target rate per modality/value
        target_rates = series_groupy_order(init_target_rates, order) / frequencies / len_df

        # identifying the underrepresented value
        discarded_idx = argmin(frequencies)
        discarded_value = order[discarded_idx]

        # choosing amongst previous and next modality (by volume and target rate)
        kept_value = find_closest_modality(
            discarded_idx,
            order,
            list(frequencies),
            list(target_rates),
            min_freq,
        )

        # removing the value from the initial ordering
        order.group(discarded_value, kept_value)

        # removing discarded_value from frequencies
        frequencies = series_groupy_order(init_frequencies, order).fillna(0)

    # case 2.2 : no underrepresented value
    return order


def series_groupy_order(series: Series, order: GroupedList) -> Series:
    """Groups a series according to groups specified in the order

    Parameters
    ----------
    series : Series
        Values to group
    order : GroupedList
        Groups of values

    Returns
    -------
    Series
        Grouped Series
    """
    grouped_series = (
        series.groupby(list(map(order.get_group, series.index)), dropna=False).sum().reindex(order)
    )

    return grouped_series


def find_closest_modality(
    idx: int, order: GroupedList, frequencies: list[float], target_rates: Series, min_freq: float
) -> Any:
    """Finds the closest modality in terms of frequency and target rate

    Parameters
    ----------
    idx : int
        _description_
    order : GroupedList
        _description_
    frequencies : list[float]
        _description_
    target_rates : Series
        _description_
    min_freq : float
        _description_

    Returns
    -------
    Any
        _description_
    """
    # case 1: lowest ranked modality
    if idx == 0:
        idx_closest_modality = 1

    # case 2: highest ranked modality
    elif idx == len(order) - 1:
        idx_closest_modality = len(order) - 2

    # case 3: modality ranked in the middle
    else:
        # previous modality's volume and target rate
        previous_freq, previous_target = frequencies[idx - 1], target_rates[idx - 1]

        # current modality's volume and target rate
        current_target = target_rates[idx]

        # next modality's volume and target rate
        next_freq, next_target = frequencies[idx + 1], target_rates[idx + 1]

        # identifying closest modality in terms of frequency
        least_frequent = idx - 1
        if next_freq < previous_freq:
            least_frequent = idx + 1

        # identifying closest modality in terms of target rate
        closest_target = idx - 1
        if abs(previous_target - current_target) >= abs(next_target - current_target):
            closest_target = idx + 1

        # case 3.1: grouping with the closest target rate
        idx_closest_modality = closest_target

        # case 3.2: (only) one modality isn't common, the least frequent is choosen
        if (previous_freq < min_freq < next_freq) or (next_freq < min_freq < previous_freq):
            idx_closest_modality = least_frequent

        # case 3.3: current modality is missing (no target rate), the least frequent is choosen
        if isna(current_target):
            idx_closest_modality = least_frequent

    # finding the closest value
    closest_value = order[idx_closest_modality]

    return closest_value



"""Base tools to build simple buckets out of Quantitative and Qualitative features
for a binary classification model.
"""

from typing import Any, Union

from numpy import array, floating, inf, integer, isfinite, nan, select
from pandas import DataFrame, Series, isna, notna, unique
from sklearn.base import BaseEstimator, TransformerMixin

from .grouped_list import GroupedList
from .serialization import json_serialize_values_orders


class BaseDiscretizer(BaseEstimator, TransformerMixin):
    """Applies discretization using a dict of GroupedList to transform a DataFrame's columns."""

    def __init__(
        self,
        features: list[str],
        *,
        values_orders: dict[str, GroupedList] = None,  #
        input_dtypes: Union[str, dict[str, str]] = "str",
        output_dtype: str = "str",
        dropna: bool = True,
        copy: bool = True,  #
        verbose: bool = True,  #
        str_nan: str = None,  #
        str_default: str = None,
        features_casting: dict[str, list[str]] = None,
    ) -> None:
        # features : list[str]
        #     List of column names of features (continuous, discrete, categorical or ordinal) to be dicretized

        # input_dtypes : Union[str, dict[str, str]], optional
        #     Input data type, converted to a dict of the provided type for each feature, by default ``"str"``

        #     * If ``"str"``, features are considered as qualitative.
        #     * If ``"float"``, features are considered as quantitative.

        # output_dtype : str, optional
        #     To be choosen amongst ``["float", "str"]``, by default ``"str"``

        #     * If ``"float"``, grouped modalities will be converted to there corresponding floating rank.
        #     * If ``"str"``, a per-group modality will be set for all the modalities of a group.

        # dropna : bool, optional
        #     * If ``True``, ``numpy.nan`` will be attributed there label.
        #     * If ``False``, ``numpy.nan`` will be restored after discretization, by default ``True``

        # str_default : str, optional
        #     String representation for default qualitative values, i.e. values less frequent than ``min_freq``, by default ``"__OTHER__"``

        # features_casting : dict[str, list[str]], optional
        #     By default ``None``, target is considered as continuous or binary.
        #     Multiclass target: Dict of raw DataFrame columns associated to the names of copies that will be created.
        """
        values_orders : dict[str, GroupedList], optional
            Dict of column names and there associated ordering.
            If lists are passed, a :class:`GroupedList` will automatically be initiated, by default ``None``

        copy : bool, optional
            If ``True``, applies transform to a copy of the provided DataFrame, by default ``False``

        verbose : bool, optional
            If ``True``, prints raw Discretizers Fit and Transform steps, by default ``False``

        str_nan : str, optional
            String representation to input ``numpy.nan``. If ``dropna=False``, ``numpy.nan`` will be left unchanged, by default ``"__NAN__"``

        Examples
        --------
        See `Discretizers examples <https://autocarver.readthedocs.io/en/latest/index.html>`_
        """
        # features and values
        self.features = list(set(features))
        if values_orders is None:
            values_orders: dict[str, GroupedList] = {}
        self.values_orders = {
            feature: GroupedList(values) for feature, values in values_orders.items()
        }
        self.copy = copy

        # input feature types
        if isinstance(input_dtypes, str):
            input_dtypes = {feature: input_dtypes for feature in features}
        self.input_dtypes = input_dtypes

        # output type
        self.output_dtype = output_dtype

        # string value of numpy.nan
        self.str_nan = str_nan

        # whether or not to reinstate numpy nan after bucketization
        self.dropna = dropna

        # string value of rare values
        self.str_default = str_default

        # whether to print info
        self.verbose = verbose

        # identifying qualitative features by there type
        self.qualitative_features = [
            feature for feature in features if self.input_dtypes[feature] == "str"
        ]

        # identifying quantitative features by there type
        self.quantitative_features = [
            feature for feature in features if self.input_dtypes[feature] == "float"
        ]

        # for each feature, getting label associated to each value
        self.labels_per_values: dict[str, dict[Any, Any]] = {}  # will be initiated during fit

        # check if the discretizer has already been fitted
        self.is_fitted = False

        # target classes multiclass vs binary vs continuous
        if features_casting is None:
            features_casting: list[Any] = {feature: [feature] for feature in self.features}
        self.features_casting = {
            feature: casting[:] for feature, casting in features_casting.items()
        }

    def _get_labels_per_values(self, output_dtype: str) -> dict[str, dict[Any, Any]]:
        """Creates a dict that contains, for each feature, for each value, the associated label

        Parameters
        ----------
        output_dtype : str
            Whether or not to convert the output to float.

        Returns
        -------
        dict[str, dict[Any, Any]]
            Dict of labels per values per feature
        """
        # initiating dict of labels per values per feature
        labels_per_values: dict[str, dict[Any, Any]] = {}

        # iterating over each feature
        for feature in self.features:
            # known values of the feature(grouped)
            values = self.values_orders[feature]

            # case 0: quantitative feature -> labels per quantile (removes str_nan)
            if feature in self.quantitative_features:
                labels = get_labels(values, self.str_nan)
            # case 1: qualitative feature -> by default, labels are values
            else:
                labels = [value for value in values if value != self.str_nan]  # (removing str_nan)

            # add NaNs if there are any
            if self.str_nan in values:
                labels += [self.str_nan]

            # requested float output (AutoCarver) -> converting to integers
            if output_dtype == "float":
                labels = [n for n, _ in enumerate(labels)]

            # building label per value
            label_per_value: dict[Any, Any] = {}
            for group_of_values, label in zip(values, labels):
                for value in values.get(group_of_values):
                    label_per_value.update({value: label})

            # storing in full dict
            labels_per_values.update({feature: label_per_value})

        return labels_per_values

    def _remove_feature(self, feature: str) -> None:
        """Removes a feature from all ``BaseDiscretizer.feature`` attributes

        Parameters
        ----------
        feature : str
            Column name of the feature to remove
        """
        if feature in self.features:
            self.features.remove(feature)
            if feature in self.qualitative_features:
                self.qualitative_features.remove(feature)
            if feature in self.quantitative_features:
                self.quantitative_features.remove(feature)
            if feature in self.values_orders:
                self.values_orders.pop(feature)
            if feature in self.input_dtypes:
                self.input_dtypes.pop(feature)
            if feature in self.labels_per_values:
                self.labels_per_values.pop(feature)

            # getting corresponding raw_feature
            raw_feature = next(
                raw_feature
                for raw_feature, casting in self.features_casting.items()
                if feature in casting
            )
            casting = self.features_casting.get(raw_feature)
            # removing feature from casting
            casting.remove(feature)
            self.features_casting.update({raw_feature: casting})
            # removing raw feature if there are no more casting for it
            if len(self.features_casting.get(raw_feature)) == 0:
                self.features_casting.pop(raw_feature)

    def _cast_features(self, X: DataFrame) -> DataFrame:
        """Casts the features of a DataFrame using features_casting to duplicate columns

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in ``BaseDiscretizer.features``, by default None.

        Returns
        -------
        DataFrame
            A formatted X
        """
        # for binary/continuous targets
        if all(len(feature_casting) == 1 for feature_casting in self.features_casting.values()):
            X.rename(
                columns={
                    feature: feature_casting[0]
                    for feature, feature_casting in self.features_casting.items()
                },
                inplace=True,
            )

        # for multiclass targets
        else:
            # duplicating features
            X = X.assign(
                **{
                    casted_feature: X[feature]
                    for feature, feature_casting in self.features_casting.items()
                    for casted_feature in feature_casting
                }
            )

        return X

    def _prepare_data(self, X: DataFrame, y: Series = None) -> DataFrame:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in ``BaseDiscretizer.features``, by default None.

        y : Series
            Binary target feature, by default None.

        Returns
        -------
        DataFrame
            A formatted copy of X
        """
        x_copy = X
        if X is not None:
            # checking for X's type
            assert isinstance(
                X, DataFrame
            ), f" - [BaseDiscretizer] X must be a pandas.DataFrame, instead {type(X)} was passed"

            # copying X
            x_copy = X
            if self.copy:
                x_copy = X.copy()

            # casting features for multiclass targets
            x_copy = self._cast_features(x_copy)

            # checking for input columns
            missing_columns = [feature for feature in self.features if feature not in x_copy]
            assert (
                len(missing_columns) == 0
            ), f" - [BaseDiscretizer] Missing features from the provided DataFrame: {str(missing_columns)}"

            if y is not None:
                # checking for y's type
                assert isinstance(
                    y, Series
                ), f" - [BaseDiscretizer] y must be a pandas.Series, instead {type(y)} was passed"

                # checking for nans in the target
                assert not any(y.isna()), " - [BaseDiscretizer] y should not contain numpy.nan"

                # checking indices
                assert all(
                    y.index == X.index
                ), f" - [BaseDiscretizer] X and y must have the same indices."

        return x_copy

    __prepare_data = _prepare_data  # private copy

    def _check_new_values(self, X: DataFrame, features: list[str]) -> None:
        """Checks for new, unexpected values, in X

        Parameters
        ----------
        X : DataFrame
            New DataFrame (at transform time)
        features : list[str]
            List of column names
        """
        # unique non-nan values in new dataframe
        uniques = X[features].apply(
            nan_unique,
            axis=0,
            result_type="expand",
        )
        uniques = applied_to_dict_list(uniques)

        # replacing unknwon values if there was a default discretization
        X.replace(
            {
                feature: {
                    val: self.str_default
                    for val in uniques[feature]
                    if val not in self.values_orders[feature].values()
                    and val != self.str_nan
                    and self.str_default in self.values_orders[feature].values()
                }
                for feature in features
            },
            inplace=True,
        )
        # updating unique non-nan values in new dataframe
        uniques = X[features].apply(
            nan_unique,
            axis=0,
            result_type="expand",
        )
        uniques = applied_to_dict_list(uniques)

        # checking for unexpected values for each feature
        for feature in features:
            # unexpected values for this feature
            unexpected = [
                val for val in uniques[feature] if val not in self.values_orders[feature].values()
            ]
            assert len(unexpected) == 0, (
                " - [BaseDiscretizer] Unexpected value! The ordering for values: "
                f"{str(list(unexpected))} of feature '{feature}' was not provided. "
                "There might be new values in your test/dev set. Consider taking a bigger "
                f"test/dev set or dropping the column {feature}."
            )

        return X

    def fit(self, X: DataFrame = None, y: Series = None) -> None:
        """Learns simple discretization of values of X according to values of y.

        Parameters
        ----------
        X : DataFrame
            Training dataset, to determine features' optimal carving
            Needs to have columns has specified in ``features`` attribute.

        y : Series
            Target with wich the association is maximized.
        """
        # checking for previous fits of the discretizer that could cause unwanted errors
        assert not self.is_fitted, (
            " - [BaseDiscretizer] This Discretizer has already been fitted. "
            "Fitting it anew could break established orders. Please initialize a new one."
        )

        # checking that all features to discretize are in values_orders
        missing_features = [
            feature for feature in self.features if feature not in self.values_orders
        ]
        assert len(missing_features) == 0, (
            " - [BaseDiscretizer] Missing values_orders for following features "
            f"{str(missing_features)}."
        )

        # for each feature, getting label associated to each value
        self.labels_per_values = self._get_labels_per_values(self.output_dtype)

        # setting fitted as True to raise alerts
        self.is_fitted = True

        return self

    def transform(self, X: DataFrame, y: Series = None) -> DataFrame:
        """Applies discretization to a DataFrame's columns.

        Parameters
        ----------
        X : DataFrame
            Dataset to be carved.
            Needs to have columns has specified in ``features`` attribute.
        y : Series, optional
            Target, by default ``None``

        Returns
        -------
        DataFrame
            Discretized X.
        """
        # * For each feature's value, the associated group label is attributed (as definid by ``values_orders``).
        # * If ``output_dtype="float"``, converts labels into floats.
        # * Data types are matched as ``input_dtypes=="str"`` for qualitative features and ``input_dtypes=="float"`` for quantitative ones.
        # * If ``copy=True``, the input DataFrame will be copied.

        # copying dataframes and casting for multiclass
        x_copy = self.__prepare_data(X, y)

        # transforming quantitative features
        if len(self.quantitative_features) > 0:
            x_copy = self._transform_quantitative(x_copy, y)

        # transforming qualitative features
        if len(self.qualitative_features) > 0:
            x_copy = self._transform_qualitative(x_copy, y)

        # reinstating nans
        if not self.dropna:
            for feature in self.features:
                label_per_value = self.labels_per_values[feature]
                if self.str_nan in label_per_value:  # checking for nans in the feature
                    x_copy[feature] = x_copy[feature].replace(label_per_value[self.str_nan], nan)

        return x_copy

    def _transform_quantitative(self, X: DataFrame, y: Series) -> DataFrame:
        """Applies discretization to a DataFrame's Quantitative columns.

        * Data types are matched as ``input_dtypes=="str"`` for qualitative features and ``input_dtypes=="float"`` for quantitative ones.
        * If ``copye=True``, the input DataFrame will be copied.

        Parameters
        ----------
        X : DataFrame
            Contains columns named after ``BaseDiscretizer.features`` attribute, by default None
        y : Series, optional
            Model target, by default None

        Returns
        -------
        DataFrame
            Discretized X.
        """
        # dataset length
        x_len = X.shape[0]

        # grouping each quantitative feature
        for feature in self.quantitative_features:
            # feature's labels associated to each quantile
            feature_values = self.values_orders[feature]

            # keeping track of nans
            nans = isna(X[feature])

            # converting nans to there corresponding quantile (if it was grouped to a quantile)
            if any(nans):
                assert feature_values.contains(
                    self.str_nan
                ), f" - [BaseDiscretizer] Unexpected value! Missing values found for feature '{feature}' at transform step but not during fit. There might be new values in your test/dev set. Consider taking a bigger test/dev set or dropping the column {feature}."
                nan_value = feature_values.get_group(self.str_nan)
                # checking that nans have been grouped to a quantile otherwise they are left as numpy.nan (for comparison purposes)
                if nan_value != self.str_nan:
                    X.loc[nans, feature] = nan_value

            # list of masks of values to replace with there respective group
            values_to_group = [
                X[feature] <= value for value in feature_values if value != self.str_nan
            ]

            # corressponding group for each value
            group_labels = [
                [self.labels_per_values[feature][value]] * x_len
                for value in feature_values
                if value != self.str_nan
            ]

            # checking for values to group
            if len(values_to_group) > 0:
                X[feature] = select(values_to_group, group_labels, default=X[feature])

            # converting nans to there value
            if any(nans):
                X.loc[nans, feature] = self.labels_per_values[feature].get(nan_value, self.str_nan)

        return X

    def _transform_qualitative(self, X: DataFrame, y: Series = None) -> DataFrame:
        """Applies discretization to a DataFrame's Qualitative columns.

        * Data types are matched as ``input_dtypes=="str"`` for qualitative features and ``input_dtypes=="float"`` for quantitative ones.
        * If ``copye=True``, the input DataFrame will be copied.

        Parameters
        ----------
        X : DataFrame
            Contains columns named after ``BaseDiscretizer.features`` attribute, by default None
        y : Series, optional
            Model target, by default None

        Returns
        -------
        DataFrame
            Discretized X.
        """
        # filling up nans from features that have some
        if self.str_nan:
            X[self.qualitative_features] = X[self.qualitative_features].fillna(self.str_nan)

        # checking that all unique values in X are in values_orders
        X = self._check_new_values(X, features=self.qualitative_features)

        # replacing values for there corresponding label
        X = X.replace(
            {
                feature: label_per_value
                for feature, label_per_value in self.labels_per_values.items()
                if feature in self.qualitative_features
            }
        )

        return X

    def to_json(self) -> str:
        """Converts to .json format.

        To be used with ``json.dump``.

        Returns
        -------
        str
            JSON serialized object
        """
        # extracting content dictionnaries
        json_serialized_base_discretizer = {
            "features": self.features,
            "values_orders": json_serialize_values_orders(self.values_orders),
            "features_casting": self.features_casting,
            "input_dtypes": self.input_dtypes,
            "output_dtype": self.output_dtype,
            "str_nan": self.str_nan,
            "str_default": self.str_default,
            "dropna": self.dropna,
            "copy": self.copy,
        }

        # dumping as json
        return json_serialized_base_discretizer

    def summary(self, feature: str = None) -> DataFrame:
        """Summarizes the data discretization process.

        By default:

            * Modality ``str_default="__OTHER__"`` is generated for features that contain non-representative modalities.
            * Modality ``str_nan="__NAN__"`` is generated for features that contain ``numpy.nan``.

        Parameters
        ----------
        feature : str, optional
            Specify for which feature to return the summary, by default ``None``

        Returns
        -------
        DataFrame
            A summary of features' values per modalities.
        """
        # storing requested feature for later
        requested_feature = None
        if feature is not None:
            requested_feature = feature[:]
            assert requested_feature in self.features, (
                f"Discretization of feature {requested_feature} was not "
                "requested or it has been dropped."
            )

        # raw label per value with output_dtype 'str'
        raw_labels_per_values = self._get_labels_per_values(output_dtype="str")

        # initiating summaries
        summaries: list[dict[str, Any]] = []
        for feature in self.features:
            # adding each value/label
            for value, label in self.labels_per_values[feature].items():
                # checking that nan where dropped
                if not (not self.dropna and value == self.str_nan):
                    # initiating feature summary (default value/label)
                    feature_summary = {
                        "feature": feature,
                        "dtype": self.input_dtypes[feature],
                        "label": label,
                        "content": value,
                    }

                    # case 0: qualitative feature -> not adding floats and integers str_default
                    if feature in self.qualitative_features:
                        if not isinstance(value, floating) and not isinstance(
                            value, float
                        ):  # checking for floats
                            if not isinstance(value, integer) and not isinstance(
                                value, int
                            ):  # checking for ints
                                if value != self.str_default:  # checking for str_default
                                    summaries += [feature_summary]

                    # case 1: quantitative feature -> take the raw label per value
                    elif feature in self.quantitative_features:
                        feature_summary.update({"content": raw_labels_per_values[feature][value]})
                        summaries += [feature_summary]

        # adding nans for quantitative features (when nan has been grouped)
        for feature in self.quantitative_features:
            # initiating feature summary (no value/label)
            feature_summary = {"feature": feature, "dtype": self.input_dtypes[feature]}
            # if there are nans -> if already added it will be dropped afterwards (unique content)
            if self.str_nan in raw_labels_per_values[feature]:
                nan_group = self.values_orders[feature].get_group(self.str_nan)
                feature_summary.update(
                    {"label": self.labels_per_values[feature][nan_group], "content": self.str_nan}
                )
                summaries += [feature_summary]

        # aggregating unique values per label
        summaries = (
            DataFrame(summaries)
            .groupby(["feature", "dtype", "label"])["content"]
            .apply(lambda u: list(unique(u)))
            .reset_index()
        )
        # sorting content
        sorted_contents: list[list[Any]] = []
        for content in summaries["content"]:
            content.sort(key=repr)
            sorted_contents += [content]
        summaries["content"] = sorted_contents
        # sorting and seting index
        summaries = summaries.sort_values(["dtype", "feature"]).set_index(["feature", "dtype"])

        # getting requested feature
        if requested_feature is not None:
            summaries = summaries.loc[requested_feature]

        return summaries


def convert_to_labels(
    features: list[str],
    quantitative_features: list[str],
    values_orders: dict[str, GroupedList],
    str_nan: str,
    dropna: bool = True,
) -> dict[str, GroupedList]:
    """Converts a values_orders values (quantiles) to labels"""
    # copying values_orders without nans
    labels_orders = {
        feature: GroupedList([value for value in values_orders[feature] if value != str_nan])
        for feature in features
    }

    # for quantitative features getting labels per quantile
    if any(quantitative_features):
        # getting group "name" per quantile
        quantiles_labels, _ = get_quantiles_labels(quantitative_features, values_orders, str_nan)

        # applying alliases to known orders
        for feature in quantitative_features:
            labels_orders.update(
                {
                    feature: GroupedList(
                        [quantiles_labels[feature][quantile] for quantile in labels_orders[feature]]
                    )
                }
            )

    # adding back nans if requested
    if not dropna:
        for feature in features:
            if str_nan in values_orders[feature]:
                order = labels_orders[feature]
                order.append(str_nan)  # adding back nans at the end of the order
                labels_orders.update({feature: order})

    return labels_orders


def convert_to_values(
    features: list[str],
    quantitative_features: list[str],
    values_orders: dict[str, GroupedList],
    label_orders: dict[str, GroupedList],
    str_nan: str,
) -> dict[str, Any]:
    """Converts a values_orders labels to values (quantiles)"""
    # for quantitative features getting labels per quantile
    if any(quantitative_features):
        # getting quantile per group "name"
        _, labels_to_quantiles = get_quantiles_labels(quantitative_features, values_orders, str_nan)

    # updating feature orders (that keeps NaNs and quantiles)
    for feature in features:
        # initial complete ordering with NAN and quantiles
        order = values_orders[feature]

        # checking for grouped modalities
        groups_to_discard = label_orders[feature].content

        # grouping the raw quantile values
        for kept_value, group_to_discard in groups_to_discard.items():
            # for qualitative features grouping as is
            # for quantitative features getting quantile per alias
            if feature in quantitative_features:
                # getting raw quantiles to be grouped
                group_to_discard = [
                    labels_to_quantiles[feature][label_discarded]
                    if label_discarded != str_nan
                    else str_nan
                    for label_discarded in group_to_discard
                ]

                # choosing the value to keep as the group
                which_to_keep = [value for value in group_to_discard if value != str_nan]
                # case 0: keeping the largest value amongst the discarded (otherwise they wont be grouped)
                if len(which_to_keep) > 0:
                    kept_value = max(which_to_keep)
                # case 1: there is only str_nan in the group (it was not grouped)
                else:
                    kept_value = group_to_discard[0]

            # grouping quantiles
            order.group_list(group_to_discard, kept_value)

        # updating ordering
        values_orders.update({feature: order})

    return values_orders


def min_value_counts(
    x: Series,
    values_orders: dict[str, GroupedList] = None,
    dropna: bool = False,
    normalize: bool = True,
) -> float:
    """Minimum of modalities' frequencies.

    Parameters
    ----------
    x : Series
        _description_
    values_orders : dict[str, GroupedList], optional
        _description_, by default None
    dropna : bool, optional
        _description_, by default False
    normalize : bool, optional
        _description_, by default True

    Returns
    -------
    float
        _description_
    """
    # modality frequency
    values = x.value_counts(dropna=dropna, normalize=normalize)

    # setting indices
    order = values_orders.get(x.name)
    if order is not None:
        values = values.reindex(order).fillna(0)

    # minimal frequency
    min_values = values.values.min()

    return min_values


def value_counts(x: Series, dropna: bool = False, normalize: bool = True) -> dict:
    """Counts the values of each modality of a series into a dictionnary

    Parameters
    ----------
    x : Series
        _description_
    dropna : bool, optional
        _description_, by default False
    normalize : bool, optional
        _description_, by default True

    Returns
    -------
    dict
        _description_
    """

    values = x.value_counts(dropna=dropna, normalize=normalize)

    return values.to_dict()


def target_rate(x: Series, y: Series, dropna: bool = True, ascending=True) -> dict:
    """Target y rate per modality of x into a dictionnary

    Parameters
    ----------
    x : Series
        _description_
    y : Series
        _description_
    dropna : bool, optional
        _description_, by default True
    ascending : bool, optional
        _description_, by default True

    Returns
    -------
    dict
        _description_
    """

    rates = y.groupby(x, dropna=dropna).mean().sort_values(ascending=ascending)

    return rates.to_dict()


def nunique(x: Series, dropna=False) -> int:
    """Computes number of unique modalities

    Parameters
    ----------
    x : Series
        _description_
    dropna : bool, optional
        _description_, by default False

    Returns
    -------
    int
        _description_
    """

    uniques = unique(x)
    n = len(uniques)

    # removing nans
    if dropna:
        if any(isna(uniques)):
            n -= 1

    return n


def get_labels(quantiles: list[float], str_nan: str) -> list[str]:
    """_summary_

    Parameters
    ----------
    feature : str
        _description_
    order : GroupedList
        _description_
    str_nan : str
        _description_

    Returns
    -------
    list[str]
        _description_
    """
    # filtering out nan and inf for formatting
    quantiles = [val for val in quantiles if val != str_nan and isfinite(val)]

    # converting quantiles in string
    labels = format_quantiles(quantiles)

    return labels


def get_quantiles_labels(
    features: list[str], values_orders: dict[str, GroupedList], str_nan: str
) -> tuple[dict[str, GroupedList], dict[str, GroupedList]]:
    """Converts a values_orders of quantiles into a values_orders of string quantiles

    Parameters
    ----------
    features : list[str]
        _description_
    values_orders : dict[str, GroupedList]
        _description_
    str_nan : str
        _description_

    Returns
    -------
    dict[str, GroupedList]
        _description_
    """
    # applying quantiles formatting to orders of specified features
    quantiles_to_labels = {}
    labels_to_quantiles = {}
    for feature in features:
        quantiles = list(values_orders[feature])
        labels = get_labels(quantiles, str_nan)

        # associates quantiles to their respective labels
        quantiles_to_labels.update(
            {feature: {quantile: alias for quantile, alias in zip(quantiles, labels)}}
        )
        # associates quantiles to their respective labels
        labels_to_quantiles.update(
            {feature: {alias: quantile for quantile, alias in zip(quantiles, labels)}}
        )

    return quantiles_to_labels, labels_to_quantiles


def format_quantiles(a_list: list[float]) -> list[str]:
    """Formats a list of float quantiles into a list of boundaries.

    Rounds quantiles to the closest power of 1000.

    Parameters
    ----------
    a_list : list[float]
        Sorted list of quantiles to convert into string

    Returns
    -------
    list[str]
        List of boundaries per quantile
    """
    # scientific formatting
    formatted_list = [f"{number:.3e}" for number in a_list]

    # stripping whitespaces
    formatted_list = [string.strip() for string in formatted_list]

    # low and high bounds per quantiles
    upper_bounds = formatted_list + [nan]
    lower_bounds = [nan] + formatted_list
    order: list[str] = []
    for lower, upper in zip(lower_bounds, upper_bounds):
        if isna(lower):
            order += [f"x <= {upper}"]
        elif isna(upper):
            order += [f"{lower} < x"]
        else:
            order += [f"{lower} < x <= {upper}"]

    return order


def nan_unique(x: Series) -> list[Any]:
    """Unique non-NaN values.

    Parameters
    ----------
    x : Series
        Values to be deduplicated.

    Returns
    -------
    list[Any]
        List of unique non-nan values
    """

    # unique values
    uniques = unique(x)

    # filtering out nans
    uniques = [u for u in uniques if notna(u)]

    return uniques


def applied_to_dict_list(applied: Union[DataFrame, Series]) -> dict[str, list[Any]]:
    """Converts a DataFrame or a List in a Dict of lists

    Parameters
    ----------
    applied : Union[DataFrame, Series]
        Result of pandas.DataFrame.apply

    Returns
    -------
    Dict[list[Any]]
        Dict of lists of rows values
    """
    # TODO: use this function whenever apply is used

    # case when it's a Series
    converted = applied.to_dict()

    # case when it's a DataFrame
    if isinstance(applied, DataFrame):
        converted = applied.to_dict(orient="list")

    return converted


def check_missing_values(
    X: DataFrame, features: list[str], known_values: dict[str, list[Any]]
) -> None:
    """Checks for missing values from X, (unexpected values in values_orders)

    Parameters
    ----------
    X : DataFrame
        New DataFrame (at transform time)
    features : list[str]
        List of column names
    known_values : dict[str, list[Any]]
        Dict of known values per column name
    """
    # unique non-nan values in new dataframe
    uniques = X[features].apply(
        nan_unique,
        axis=0,
        result_type="expand",
    )
    uniques = applied_to_dict_list(uniques)

    # checking for unexpected values for each feature
    for feature in features:
        unexpected = [val for val in known_values[feature] if val not in uniques[feature]]
        assert (
            len(unexpected) == 0
        ), f"Unexpected value! The ordering for values: {str(list(unexpected))} of feature '{feature}' was provided but there are not matching value in provided X. You should check 'values_orders['{feature}']' for unwanted values."


class extend_docstring:
    """Used to extend a Child's method docstring with the Parent's method docstring"""

    def __init__(self, method):
        self.doc = method.__doc__

    def __call__(self, function):
        if self.doc is not None:
            doc = function.__doc__
            function.__doc__ = self.doc
            if doc is not None:
                function.__doc__ = doc + function.__doc__
        return function



""" Measures of association between a Qualitative feature and binary target.
"""

from math import sqrt
from typing import Any, Callable

from pandas import Series, crosstab, notna
from scipy.stats import chi2_contingency

# X continue y continue distance correlation
# X binaire y continue kruskal y, x
# X multiclass y continue kruskal y, x

# X continue y binaire kruskal x, y
# X binaire y binaire cramerv/tschurpowt x, y
# X multiclass y binaire cramerv/tschurpowt x, y


# X continue y multiclass kruskal x, y
# X binaire y multiclass cramerv/tschurpowt x, y
# X multiclass y multiclass cramerv/tschurpowt x, y


def chi2_measure(
    x: Series,
    y: Series,
    thresh_chi2: float = 0,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Wrapper for `scipy.stats.chi2_contingency <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.chi2_contingency.html>`_.
    Computes Chi2 statistic on the ``x`` by ``y`` `pandas.crosstab <https://pandas.pydata.org/docs/reference/api/pandas.crosstab.html>`_.

    Parameters
    ----------
    x : Series
        Qualitative feature
    y : Series
        Qualitative target feature
    thresh_chi2 : float, optional
        Minimum Chi2 association, by default ``0``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether ``x`` is sufficiently associated to ``y`` and Pearson's chi2 between ``x`` and ``y``.
    """
    # computing crosstab between x and y
    xtab = crosstab(x, y)

    # Chi2 statistic
    chi2 = chi2_contingency(xtab)[0]

    # updating association
    measurement = {"chi2_statistic": chi2}

    # Excluding features not associated enough
    active = chi2 < thresh_chi2

    return active, measurement


def cramerv_measure(
    x: Series,
    y: Series,
    thresh_cramerv: float = 0,
    chi2_statistic: float = None,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Computes Carmér's V between ``x`` and ``y`` from ``chi2_measure``.

    Parameters
    ----------
    x : Series
        Qualitative feature
    y : Series
        Qualitative target feature
    thresh_cramerv : float, optional
        Minimum Cramér's V association, by default ``0``
    chi2_statistic : float, optional
        Pearson's chi2 between ``x`` and ``y``, by default ``None``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether ``x`` is sufficiently associated to ``y`` and Carmér's V between ``x`` and ``y``.
    """
    # Chi2 statistic
    if chi2_statistic is None:
        _, measurement = chi2_measure(x, y, **kwargs)
        chi2_statistic = measurement.get("chi2_statistic")

    # number of observations
    n_obs = (notna(x) & notna(y)).sum()

    # number of values taken by the features
    n_mod_x, n_mod_y = x.nunique(), y.nunique()
    min_n_mod = min(n_mod_x, n_mod_y)

    # Cramér's V
    cramerv = sqrt(chi2_statistic / n_obs / (min_n_mod - 1))

    # updating association
    measurement.update({"cramerv_measure": cramerv})

    # Excluding features not associated enough
    active = cramerv < thresh_cramerv

    return active, measurement


def tschuprowt_measure(
    x: Series,
    y: Series,
    thresh_tschuprowt: float = 0,
    chi2_statistic: float = None,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Computes Tschuprow's T between ``x`` and ``y`` from ``chi2_measure``.

    Parameters
    ----------
    x : Series
        Feature to measure
    y : Series
        Binary target feature
    thresh_tschuprowt : float, optional
        Minimum Tschuprow's T association, by default ``0``
    chi2_statistic : float, optional
        Pearson's chi2 between ``x`` and ``y``, by default ``None``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether ``x`` is sufficiently associated to ``y`` and Tschuprow's T between ``x`` and ``y``.
    """
    # Chi2 statistic
    if chi2_statistic is None:
        _, measurement = chi2_measure(x, y, **kwargs)
        chi2_statistic = measurement.get("chi2_statistic")

    # number of observations
    n_obs = (notna(x) & notna(y)).sum()

    # number of values taken by the features
    n_mod_x, n_mod_y = x.nunique(), y.nunique()

    # Tschuprow's T
    dof_mods = sqrt((n_mod_x - 1) * (n_mod_y - 1))
    tschuprowt = 0
    if dof_mods > 0:
        tschuprowt = sqrt(chi2_statistic / n_obs / dof_mods)

    # updating association
    measurement.update({"tschuprowt_measure": tschuprowt})

    # Excluding features not associated enough
    active = tschuprowt < thresh_tschuprowt

    return active, measurement



""" Loads association filters."""

from ...selectors.filters.base_filters import thresh_filter
from ...selectors.filters.qualitative_filters import cramerv_filter, tschuprowt_filter
from ...selectors.filters.quantitative_filters import pearson_filter, spearman_filter



"""(For compatibility) Tools to select the best Quantitative and Qualitative features."""

from typing import Callable
from warnings import warn

from ..discretizers import GroupedList, extend_docstring
from ..selectors import ClassificationSelector


class FeatureSelector(ClassificationSelector):
    @extend_docstring(ClassificationSelector.__init__)
    def __init__(
        self,
        n_best: int,
        qualitative_features: list[str] = None,
        quantitative_features: list[str] = None,
        *,
        measures: list[Callable] = None,
        filters: list[Callable] = None,
        colsample: float = 1.0,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        warn(
            "feature_selection.FeatureSelector will be deprecated, please use selectors.ClassificationSelector instead.",
            DeprecationWarning,
        )
        # corresponding measures and filters
        quantitative_measures = None
        qualitative_measures = None
        if measures is not None:
            if (quantitative_features is not None) and any(quantitative_features):
                quantitative_measures = measures[:]
            elif (qualitative_features is not None) and any(qualitative_features):
                qualitative_measures = measures[:]

        quantitative_filters = None
        qualitative_filters = None
        if filters is not None:
            if (quantitative_features is not None) and any(quantitative_features):
                quantitative_filters = filters[:]
            elif (qualitative_features is not None) and any(qualitative_features):
                qualitative_filters = filters[:]

        # initiating BaseSelector with the corresponding list of measures
        super().__init__(
            n_best=n_best,
            quantitative_features=quantitative_features,
            qualitative_features=qualitative_features,
            quantitative_measures=quantitative_measures,
            qualitative_measures=qualitative_measures,
            quantitative_filters=quantitative_filters,
            qualitative_filters=qualitative_filters,
            colsample=colsample,
            verbose=verbose,
            **kwargs,
        )



""" Filters based on association measures between Qualitative features.
"""

from typing import Any, Callable

from pandas import DataFrame

from ..measures import cramerv_measure, make_measure, tschuprowt_measure


def cramerv_filter(
    X: DataFrame, ranks: DataFrame, thresh_corr: float = 1, **kwargs
) -> dict[str, Any]:
    """Computes maximum Cramer's V between ``X`` and ``X`` (qualitative).
    Features too correlated to a feature more associated with the target
    are excluded (according to provided ``ranks``).

    Parameters
    ----------
    X : DataFrame
        Contains columns named after ``ranks``'s index (feature names)
    ranks : DataFrame
        Ranked features as index of the association table
    thresh_corr : float, optional
        Maximum Cramér's V bewteen features, by default ``1``

    Returns
    -------
    dict[str, Any]
        Maximum Cramér's V with a better feature
    """

    # applying quantitative filter with Cramer's V correlation
    return qualitative_filter(X, ranks, cramerv_measure, thresh_corr, **kwargs)


def tschuprowt_filter(
    X: DataFrame, ranks: DataFrame, thresh_corr: float = 1, **kwargs
) -> dict[str, Any]:
    """Computes max Tschuprow's T between X and X (qualitative).
    Features too correlated to a feature more associated with the target
    are excluded (according to provided ``ranks``).

    Parameters
    ----------
    X : DataFrame
        Contains columns named after ``ranks``'s index (feature names)
    ranks : DataFrame
        Ranked features as index of the association table
    thresh_corr : float, optional
        Maximum Tschuprow's T bewteen features, by default ``1``

    Returns
    -------
    dict[str, Any]
        Maximum Tschuprow's T with a better feature
    """

    # applying quantitative filter with Tschuprow's T correlation
    return qualitative_filter(X, ranks, tschuprowt_measure, thresh_corr, **kwargs)


def qualitative_filter(
    X: DataFrame, ranks: DataFrame, corr_measure: Callable, thresh_corr: float = 1, **kwargs
) -> dict[str, Any]:
    """Computes max association between X and X (qualitative) excluding features
    that are correlated to a feature more associated with the target
    (defined by the ranks).
    """

    # accessing the prefered order
    prefered_order = ranks.index

    # initiating list of maximum association per feature
    associations = []

    # iterating over each feature by target association order
    for feature in prefered_order:
        # computing correlation with better features anf filtering out ranks
        ranks, worst_corr = qualitative_worst_corr(
            X, feature, ranks, corr_measure, thresh_corr, **kwargs
        )

        # updating associations
        if worst_corr:
            associations += [worst_corr]

    # checking for some selected features
    if len(associations) > 0:
        # formatting ouput to DataFrame
        associations = DataFrame(associations).set_index("feature")

        # applying filter on association
        associations = ranks.join(associations, how="right")

    return associations


def qualitative_worst_corr(
    X: DataFrame,
    feature: str,
    ranks: DataFrame,
    corr_measure: Callable,
    thresh_corr: float,
    **kwargs,
):
    """Computes maximum association between a feature and features
    more associated to the target (according to ranks)
    """
    # measure name
    measure_name = corr_measure.__name__
    measure = measure_name.replace("_measure", "")

    # initiating worst correlation
    worst_corr = {"feature": feature}

    # features more associated with target
    better_features = list(ranks.loc[:feature].index)[:-1]

    # iterating over each better feature
    for better_feature in better_features:
        # computing association with better feature
        _, association = make_measure(
            corr_measure,
            True,
            {f"{measure}_with": better_feature},
            X[feature],
            X[better_feature],
            **kwargs,
        )

        # updating association if it's greater than previous better features
        if association.get(measure_name) > worst_corr.get(measure_name, 0):
            # renaming association measure as filter
            association[f"{measure}_filter"] = association.pop(measure_name)

            # removing temporary measures
            association = {k: v for k, v in association.items() if "_statistic" not in k}

            # updating worst known association
            worst_corr.update(association)

        # stopping measurements if association is greater than threshold
        if association.get(f"{measure}_filter") > thresh_corr:
            ranks = ranks.drop(feature, axis=0)  # removing feature from ranks

            return ranks, None

    return ranks, worst_corr



"""Tool to build optimized buckets out of Quantitative and Qualitative features
for binary classification tasks.
"""

from typing import Callable
from warnings import warn

from numpy import add, array, searchsorted, sqrt, unique, zeros
from pandas import DataFrame, Series, crosstab
from scipy.stats import chi2_contingency

from AutoCarver.discretizers import GroupedList

from ..discretizers import GroupedList
from ..discretizers.utils.base_discretizers import extend_docstring
from .base_carver import BaseCarver


class BinaryCarver(BaseCarver):
    """Automatic carving of continuous, discrete, categorical and ordinal
    features that maximizes association with a binary target.
    """

    @extend_docstring(BaseCarver.__init__)
    def __init__(
        self,
        sort_by: str,
        min_freq: float,
        *,
        quantitative_features: list[str] = None,
        qualitative_features: list[str] = None,
        ordinal_features: list[str] = None,
        values_orders: dict[str, GroupedList] = None,
        max_n_mod: int = 5,
        min_freq_mod: float = None,
        output_dtype: str = "float",
        dropna: bool = True,
        copy: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        sort_by : str
            Metric to be used to perform association measure between features and target.

            * ``"tschuprowt"``, for :ref:`Tschuprowt`.
            * ``"cramerv"``, for :ref:`Cramerv`.

            **Tip**: use ``"tschuprowt"`` for more robust, or less output modalities,
            use ``"cramerv"`` for more output modalities.
        """
        # association measure used to find the best groups for binary targets
        implemented_measures = ["tschuprowt", "cramerv"]
        assert sort_by in implemented_measures, (
            f" - [BinaryCarver] Measure '{sort_by}' not yet implemented for binary targets. "
            f"Choose from: {str(implemented_measures)}."
        )

        # Initiating BaseCarver
        super().__init__(
            min_freq=min_freq,
            sort_by=sort_by,
            quantitative_features=quantitative_features,
            qualitative_features=qualitative_features,
            ordinal_features=ordinal_features,
            values_orders=values_orders,
            max_n_mod=max_n_mod,
            min_freq_mod=min_freq_mod,
            output_dtype=output_dtype,
            dropna=dropna,
            copy=copy,
            verbose=verbose,
            **kwargs,
        )

    def _prepare_data(
        self,
        X: DataFrame,
        y: Series,
        X_dev: DataFrame = None,
        y_dev: Series = None,
    ) -> tuple[DataFrame, DataFrame, dict[str, Callable]]:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in ``AutoCarver.features``.

        y : Series
            Binary target feature with wich the association is maximized.

        X_dev : DataFrame, optional
            Dataset to evalute the robustness of discretization, by default ``None``
            It should have the same distribution as X.

        y_dev : Series, optional
            Binary target feature with wich the robustness of discretization is evaluated, by default ``None``

        Returns
        -------
        tuple[DataFrame, DataFrame, dict[str, Callable]]
            Copies of (X, X_dev) and helpers to be used according to target type
        """
        # Checking for binary target and copying X
        x_copy, x_dev_copy = super()._prepare_data(X, y, X_dev=X_dev, y_dev=y_dev)

        # binary target, checking values
        y_values = unique(y)
        assert (0 in y_values) & (
            1 in y_values
        ), " - [BinaryCarver] y must be a binary Series (int or float, not object)"
        assert (
            len(y_values) == 2
        ), " - [BinaryCarver] y must be a binary Series (int or float, not object)"

        return x_copy, x_dev_copy

    def _aggregator(
        self, features: list[str], X: DataFrame, y: Series, labels_orders: dict[str, GroupedList]
    ) -> dict[str, DataFrame]:
        """Computes crosstabs for specified features and ensures that the crosstab is ordered according to the known labels

        Parameters
        ----------
        features : list[str]
            _description_
        X : DataFrame
            _description_
        y : Series
            _description_
        labels_orders : dict[str, GroupedList]
            _description_

        Returns
        -------
        dict[str, DataFrame]
            _description_
        """
        # checking for empty datasets
        xtabs = {feature: None for feature in features}
        if X is not None:
            # crosstab for each feature
            for feature in features:
                # computing crosstab with str_nan
                xtab = crosstab(X[feature], y)

                # reordering according to known_order
                xtab = xtab.reindex(labels_orders[feature])

                # storing results
                xtabs.update({feature: xtab})

        return xtabs

    def _grouper(self, xtab: DataFrame, groupby: list[str]) -> DataFrame:
        """Groups a crosstab by groupby and sums column values by groups (vectorized)

        Parameters
        ----------
        xtab : DataFrame
            _description_
        groupby : list[str]
            _description_

        Returns
        -------
        DataFrame
            _description_
        """
        # all indices that may be duplicated
        index_values = array([groupby.get(index_value, index_value) for index_value in xtab.index])

        # all unique indices deduplicated
        unique_indices = unique(index_values)

        # initiating summed up array with zeros
        summed_values = zeros((len(unique_indices), len(xtab.columns)))

        # for each unique_index found in index_values sums xtab.Values at corresponding position in summed_values
        add.at(summed_values, searchsorted(unique_indices, index_values), xtab.values)

        # converting back to dataframe
        return DataFrame(summed_values, index=unique_indices, columns=xtab.columns)

    def _association_measure(self, xtab: DataFrame, n_obs: int) -> dict[str, float]:
        """Computes measures of association between feature and target by crosstab.

        Parameters
        ----------
        xtab : DataFrame
            Crosstab between feature and target.

        n_obs : int
            Sample total size.

        Returns
        -------
        dict[str, float]
            Cramér's V and Tschuprow's as a dict.
        """
        # number of values taken by the features
        n_mod_x = xtab.shape[0]

        # Chi2 statistic
        chi2 = chi2_contingency(xtab)[0]

        # Cramér's V
        cramerv = sqrt(chi2 / n_obs)

        # Tschuprow's T
        tschuprowt = cramerv / sqrt(sqrt(n_mod_x - 1))

        return {"cramerv": cramerv, "tschuprowt": tschuprowt}

    def _target_rate(self, xtab: DataFrame) -> DataFrame:
        """Computes target rate per row for a binary target (column) in a crosstab

        Parameters
        ----------
        xtab : DataFrame
            _description_

        Returns
        -------
        DataFrame
            _description_
        """
        return xtab[1].divide(xtab.sum(axis=1)).sort_values()

    def _printer(self, xtab: DataFrame = None) -> DataFrame:
        """Prints a binary xtab's statistics

        Parameters
        ----------
        xtab : Dataframe
            A crosstab, by default None

        Returns
        -------
        DataFrame
            Target rate and frequency per modality
        """
        # checking for an xtab
        stats = None
        if xtab is not None:
            # target rate and frequency statistics per modality
            stats = DataFrame(
                {
                    # target rate per modality
                    "target_rate": xtab[1].divide(xtab.sum(axis=1)),
                    # frequency per modality
                    "frequency": xtab.sum(axis=1) / xtab.sum().sum(),
                }
            )

        return stats

    @extend_docstring(BaseCarver.fit)
    def fit(
        self,
        X: DataFrame,
        y: Series,
        *,
        X_dev: DataFrame = None,
        y_dev: Series = None,
    ) -> None:
        # preparing datasets and checking for wrong values
        x_copy, x_dev_copy = self._prepare_data(X, y, X_dev, y_dev)

        # Fitting BaseCarver
        super().fit(x_copy, y, X_dev=x_dev_copy, y_dev=y_dev)

        return self


class AutoCarver(BinaryCarver):
    @extend_docstring(BinaryCarver.__init__)
    def __init__(
        self,
        sort_by: str,
        min_freq: float,
        *,
        quantitative_features: list[str] = None,
        qualitative_features: list[str] = None,
        ordinal_features: list[str] = None,
        values_orders: dict[str, GroupedList] = None,
        max_n_mod: int = 5,
        output_dtype: str = "float",
        dropna: bool = True,
        copy: bool = False,
        verbose: bool = False,
        pretty_print: bool = False,
        **kwargs,
    ) -> None:
        warn("AutoCarver will be deprecated, please use BinaryCarver instead.", DeprecationWarning)
        super().__init__(
            sort_by,
            min_freq,
            quantitative_features=quantitative_features,
            qualitative_features=qualitative_features,
            ordinal_features=ordinal_features,
            values_orders=values_orders,
            max_n_mod=max_n_mod,
            output_dtype=output_dtype,
            dropna=dropna,
            copy=copy,
            verbose=verbose,
            pretty_print=pretty_print,
            **kwargs,
        )



"""Tool to build optimized buckets out of Quantitative and Qualitative features
for any task.
"""

from typing import Any, Union
from warnings import warn

from numpy import isclose
from pandas import DataFrame, Series, unique
from tqdm import tqdm

from ..discretizers import GroupedList
from ..discretizers.discretizers import Discretizer
from ..discretizers.utils.base_discretizers import (
    BaseDiscretizer,
    convert_to_labels,
    convert_to_values,
)
from ..discretizers.utils.serialization import json_deserialize_values_orders

# trying to import extra dependencies
try:
    from IPython.display import display_html
except ImportError:
    _has_idisplay = False
else:
    _has_idisplay = True


class BaseCarver(BaseDiscretizer):
    """Automatic carving of continuous, discrete, categorical and ordinal
    features that maximizes association with a binary or continuous target.

    First fits a :class:`Discretizer`. Raw data should be provided as input (not a result of ``Discretizer.transform()``).
    """

    def __init__(
        self,
        sort_by: str,
        min_freq: float,
        *,
        quantitative_features: list[str] = None,
        qualitative_features: list[str] = None,
        ordinal_features: list[str] = None,
        values_orders: dict[str, GroupedList] = None,
        max_n_mod: int = 5,
        min_freq_mod: float = None,
        output_dtype: str = "float",
        dropna: bool = True,
        copy: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        """
        min_freq : float
            Minimum frequency per grouped modalities.

            * Features whose most frequent modality is less frequent than ``min_freq`` will not be carved.
            * Sets the number of quantiles in which to discretize the continuous features.
            * Sets the minimum frequency of a quantitative feature's modality.

            **Tip**: should be set between ``0.01`` (slower, preciser, less robust) and ``0.2`` (faster, more robust)

        quantitative_features : list[str], optional
            List of column names of quantitative features (continuous and discrete) to be carved, by default ``None``

        qualitative_features : list[str], optional
            List of column names of qualitative features (non-ordinal) to be carved, by default ``None``

        ordinal_features : list[str], optional
            List of column names of ordinal features to be carved. For those features a list of
            values has to be provided in the ``values_orders`` dict, by default ``None``

        values_orders : dict[str, GroupedList], optional
            Dict of feature's column names and there associated ordering.
            If lists are passed, a :class:`GroupedList` will automatically be initiated, by default ``None``

        max_n_mod : int, optional
            Maximum number of modality per feature, by default ``5``

            All combinations of modalities for groups of modalities of sizes from 1 to ``max_n_mod`` will be tested.
            The combination with the best association will be selected.

            **Tip**: should be set between ``3`` (faster, more robust) and ``7`` (slower, preciser, less robust)

        min_freq_mod : float, optional
            Minimum frequency per final modality, by default ``None`` for ``min_freq/2``

        output_dtype : str, optional
            To be choosen amongst ``["float", "str"]``, by default ``"float"``

            * ``"float"``, sets the rank of modalities as label.
            * ``"str"``, sets one modality of group as label.

        dropna : bool, optional
            * ``True``, try to group ``numpy.nan`` with other modalities.
            * ``False``, ``numpy.nan`` are ignored (not grouped), by default ``True``

        copy : bool, optional
            If ``True``, feature processing at transform is applied to a copy of the provided DataFrame, by default ``False``

        verbose : bool, optional
            * ``True``, without IPython installed: prints raw Discretizers and AutoCarver Fit steps for X, by default ``False``
            * ``True``, with IPython installed: adds HTML tables of target rates and frequencies for X and X_dev.

            **Tip**: IPython displaying can be turned off by setting ``pretty_print=False``.

        **kwargs
            Pass values for ``str_default`` and ``str_nan`` (default string values),
            as long as ``pretty_print`` to turn off IPython.

        Examples
        --------
        See `AutoCarver examples <https://autocarver.readthedocs.io/en/latest/index.html>`_
        """
        # Lists of features
        if quantitative_features is None:
            quantitative_features = []
        if qualitative_features is None:
            qualitative_features = []
        if ordinal_features is None:
            ordinal_features = []
        assert (
            len(quantitative_features) > 0
            or len(qualitative_features) > 0
            or len(ordinal_features) > 0
        ), (
            " - [AutoCarver] No feature passed as input. Pleased provided column names to Carver "
            "by setting quantitative_features, quantitative_features or ordinal_features."
        )
        self.ordinal_features = list(set(ordinal_features))
        self.features = list(set(quantitative_features + qualitative_features + ordinal_features))

        # initializing input_dtypes
        self.input_dtypes = {feature: "str" for feature in qualitative_features + ordinal_features}
        self.input_dtypes.update({feature: "float" for feature in quantitative_features})

        # Initiating BaseDiscretizer
        super().__init__(
            features=self.features,
            values_orders=values_orders,
            input_dtypes=self.input_dtypes,
            output_dtype=output_dtype,
            str_nan=kwargs.get("str_nan", "__NAN__"),
            str_default=kwargs.get("str_default", "__OTHER__"),
            dropna=dropna,
            copy=copy,
            verbose=bool(max(verbose, kwargs.get("pretty_print", False))),
        )

        # checking that qualitatitve and quantitative features are distinct
        assert all(
            quali_feature not in self.quantitative_features
            for quali_feature in self.qualitative_features
        ), (
            " - [AutoCarver] One of quantitative_features is also in qualitative_features "
            "or ordinal_features. Please, be carreful with your inputs!"
        )
        assert all(
            quanti_feature not in self.qualitative_features
            for quanti_feature in self.quantitative_features
        ), (
            " - [AutoCarver] One of qualitative_features or ordinal_features is also in "
            "quantitative_features. Please, be carreful with your inputs!"
        )

        # class specific attributes
        self.min_freq = min_freq  # minimum frequency per base bucket
        self.max_n_mod = max_n_mod  # maximum number of modality per feature
        if min_freq_mod is None:
            min_freq_mod = min_freq / 2
        self.min_freq_mod = min_freq_mod  # minimum frequency per final bucket
        self.sort_by = sort_by
        self.pretty_print = False
        if self.verbose and kwargs.get("pretty_print", True):
            if _has_idisplay:  # checking for installed dependencies
                self.pretty_print = True
            else:
                warn(
                    "Package not found: IPython. Defaulting to raw verbose. "
                    "Install extra dependencies with pip install autocarver[jupyter]",
                    UserWarning,
                )

        # historizing everything
        self._history = {feature: [] for feature in self.features}

    def _prepare_data(
        self,
        X: DataFrame,
        y: Series,
        X_dev: DataFrame = None,
        y_dev: Series = None,
    ) -> tuple[DataFrame, DataFrame]:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in ``BaseCarver.features``.

        y : Series
            Binary target feature with wich the association is maximized.

        X_dev : DataFrame, optional
            Dataset to evalute the robustness of discretization, by default ``None``
            It should have the same distribution as X.

        y_dev : Series, optional
            Binary target feature with wich the robustness of discretization is evaluated, by default ``None``

        Returns
        -------
        tuple[DataFrame, DataFrame]
            Copies of (X, X_dev)
        """
        # Checking for binary target and copying X
        x_copy = super()._prepare_data(X, y)
        x_dev_copy = super()._prepare_data(X_dev, y_dev)

        # checking for not provided y
        assert y is not None, f" - [AutoCarver] y must be provided {y}"

        return x_copy, x_dev_copy

    def _combination_formatter(self, combination: list[list[str]]) -> dict[str, str]:
        """Attributes the first element of a group to all elements of a group

        Parameters
        ----------
        combination : list[list[str]]
            _description_

        Returns
        -------
        dict[str, str]
            _description_
        """
        return {modal: group[0] for group in combination for modal in group}

    def _remove_feature(self, feature: str) -> None:
        """Removes a feature from all ``BaseCarver.features`` attributes

        Parameters
        ----------
        feature : str
            Column name of the feature to remove
        """
        if feature in self.features:
            super()._remove_feature(feature)
            if feature in self.ordinal_features:
                self.ordinal_features.remove(feature)
            if feature in self._history:
                self._history[feature] += [{"removed": True}]

    def fit(
        self,
        X: DataFrame,
        y: Series,
        *,
        X_dev: DataFrame = None,
        y_dev: Series = None,
    ) -> None:
        """Finds the combination of modalities of X that provides the best association with y.

        Parameters
        ----------
        X : DataFrame
            Training dataset, to determine features' optimal carving.
            Needs to have columns has specified in ``features`` attribute.

        y : Series
            Target with wich the association is maximized.

        X_dev : DataFrame, optional
            Development dataset, to evaluate robustness of carved features, by default ``None``
            Should have the same distribution as X.

        y_dev : Series, optional
            Target of the development dataset, by default ``None``
            Should have the same distribution as y.
        """
        # preparing datasets and checking for wrong values
        x_copy, x_dev_copy = self._prepare_data(X, y, X_dev, y_dev)

        # discretizing all features
        discretizer = Discretizer(
            quantitative_features=self.quantitative_features,
            qualitative_features=self.qualitative_features,
            min_freq=self.min_freq,
            ordinal_features=self.ordinal_features,
            values_orders=self.values_orders,
            str_nan=self.str_nan,
            str_default=self.str_default,
            copy=True,  # copying anyways, otherwise no discretization from start to finish
            verbose=self.verbose,
        )
        x_copy = discretizer.fit_transform(x_copy, y)
        if x_dev_copy is not None:
            x_dev_copy = discretizer.transform(x_dev_copy, y_dev)
        self.input_dtypes.update(discretizer.input_dtypes)  # saving data types

        # updating values_orders according to base bucketization
        self.values_orders.update(discretizer.values_orders)

        # removing dropped features
        removed_features = [
            feature for feature in self.features if feature not in discretizer.features
        ]
        for feature in removed_features:
            self._remove_feature(feature)

        # converting quantiles into there respective labels
        labels_orders = convert_to_labels(
            features=self.features,
            quantitative_features=self.quantitative_features,
            values_orders=self.values_orders,
            str_nan=self.str_nan,
            dropna=False,
        )

        # computing crosstabs for each feature on train/test
        xaggs = self._aggregator(self.features, x_copy, y, labels_orders)
        xaggs_dev = self._aggregator(self.features, x_dev_copy, y_dev, labels_orders)

        # optimal butcketization/carving of each feature
        all_features = self.features[:]  # (features are being removed from self.features)
        for n, feature in enumerate(all_features):
            if self.verbose:  # verbose if requested
                print(f"\n------\n[AutoCarver] Fit {feature} ({n+1}/{len(all_features)})\n---")

            # carving the feature
            labels_orders = self._carve_feature(feature, xaggs, xaggs_dev, labels_orders)

            if self.verbose:  # verbose if requested
                print("------\n")

        # discretizing features based on each feature's values_order
        super().fit(X, y)

        return self

    def _carve_feature(
        self,
        feature: str,
        xaggs: dict[str, Union[Series, DataFrame]],
        xaggs_dev: dict[str, Union[Series, DataFrame]],
        labels_orders: dict[str, GroupedList],
    ) -> dict[str, GroupedList]:
        """Carves a feature into buckets that maximize association with the target

        Parameters
        ----------
        feature : str
            _description_
        xaggs : dict[str, Union[Series, DataFrame]]
            _description_
        xaggs_dev : dict[str, Union[Series, DataFrame]]
            _description_
        labels_orders : dict[str, GroupedList]
            _description_

        Returns
        -------
        dict[str, GroupedList]
            _description_
        """
        # getting xtabs on train/test
        xagg = xaggs[feature]
        xagg_dev = xaggs_dev[feature]

        # checking that the feature has at least 2 modalities
        best_combination = None
        if len(xagg.index) > 1:
            # ordering
            order = labels_orders[feature]

            # historizing raw combination
            raw_association = {
                "index_to_groupby": {modality: modality for modality in xagg.index},
                self.sort_by: self._association_measure(xagg, n_obs=xagg.apply(sum).sum())[
                    self.sort_by
                ],
            }
            self._historize_viability_test(feature, raw_association, order)

            # verbose
            self._print_xagg(
                xagg=self._index_mapper(feature, labels_orders, xagg),
                xagg_dev=self._index_mapper(feature, labels_orders, xagg_dev),
                message="Raw distribution",
            )

            # getting best combination
            best_combination = self._get_best_combination(feature, order, xagg, xagg_dev=xagg_dev)

        # checking that a suitable combination has been found
        if best_combination is not None:
            order, xagg, xagg_dev = best_combination
            # updating orders accordingly
            labels_orders = self._update_orders(feature, order, labels_orders)

            # verbose
            self._print_xagg(
                xagg=self._index_mapper(feature, labels_orders, xagg),
                xagg_dev=self._index_mapper(feature, labels_orders, xagg_dev),
                message="Carved distribution",
            )

        # no suitable combination has been found -> removing feature
        else:
            warn(
                f" - [AutoCarver] No robust combination for feature '{feature}' could be found"
                ". It will be ignored. You might have to increase the size of your dev sample"
                " (dev sample not representative of dev sample for this feature) or you"
                " should consider dropping this features.",
                UserWarning,
            )
            self._remove_feature(feature)
            if feature in labels_orders:
                labels_orders.pop(feature)

        return labels_orders

    def _update_orders(
        self, feature: str, new_order: GroupedList, labels_orders: dict[str, GroupedList]
    ) -> dict[str, GroupedList]:
        """updates values_orders and labels_orders accoding to the new order for specified feature"""

        # updating label_orders
        labels_orders.update({feature: new_order})

        # updating values_orders
        self.values_orders.update(
            convert_to_values(
                features=[feature],
                quantitative_features=[feature] if feature in self.quantitative_features else [],
                values_orders=self.values_orders,
                label_orders=labels_orders,
                str_nan=self.str_nan,
            )
        )
        # updating labels
        labels_orders = convert_to_labels(
            features=self.features,
            quantitative_features=self.quantitative_features,
            values_orders=self.values_orders,
            str_nan=self.str_nan,
            dropna=False,
        )

        return labels_orders

    def _get_best_combination(
        self,
        feature: str,
        order: GroupedList,
        xagg: DataFrame,
        *,
        xagg_dev: DataFrame = None,
    ) -> tuple[GroupedList, DataFrame, DataFrame]:
        """_summary_

        Parameters
        ----------
        order : GroupedList
            _description_
        xagg : DataFrame
            _description_
        xagg_dev : DataFrame, optional
            _description_, by default None

        Returns
        -------
        tuple[GroupedList, DataFrame, DataFrame]
            _description_
        """
        # raw ordering
        raw_order = GroupedList(order)
        if self.str_nan in raw_order:
            raw_order.remove(self.str_nan)

        # filtering out nans if requested from train/test crosstabs
        raw_xagg = filter_nan(xagg, self.str_nan)
        raw_xagg_dev = filter_nan(xagg_dev, self.str_nan)

        # checking for non-nan values
        best_association = None
        if raw_xagg.shape[0] > 1:
            # all possible consecutive combinations
            combinations = consecutive_combinations(raw_order, self.max_n_mod, min_group_size=1)

            # getting most associated combination
            best_association, order = self._get_best_association(
                feature,
                order,
                raw_xagg,
                combinations,
                xagg_dev=raw_xagg_dev,
            )

            # applying best_combination to order and xtabs
            if best_association is not None:
                xagg = xagg_apply_order(xagg, order)
                xagg_dev = xagg_apply_order(xagg_dev, order)

            # grouping NaNs if requested to drop them (dropna=True)
            if self.dropna and self.str_nan in order and best_association is not None:
                # raw ordering without nans
                raw_order = GroupedList(order)
                raw_order.remove(self.str_nan)

                # adding combinations with NaNs
                combinations = nan_combinations(raw_order, self.str_nan, self.max_n_mod, 1)

                # getting most associated combination
                best_association, order = self._get_best_association(
                    feature,
                    order,
                    xagg,
                    combinations,
                    xagg_dev=xagg_dev,
                    dropna=True,
                )

                # applying best_combination to order and xtab
                if best_association is not None:
                    xagg = xagg_apply_order(xagg, order)
                    xagg_dev = xagg_apply_order(xagg_dev, order)

        # checking that a suitable combination has been found
        if best_association is not None:
            return order, xagg, xagg_dev

    def _get_best_association(
        self,
        feature: str,
        order: GroupedList,
        xagg: Union[Series, DataFrame],
        combinations: list[list[str]],
        *,
        xagg_dev: Union[Series, DataFrame] = None,
        dropna: bool = False,
    ) -> tuple[dict[str, Any], GroupedList]:
        """Computes associations of the tab for each combination

        Parameters
        ----------
        feature : str
            feature to measure association with
        xagg : Union[Series, DataFrame]
            _description_
        order : GroupedList
            order of the feature
        combinations : list[list[str]]
            _description_
        xagg_dev : Union[Series, DataFrame], optional
            _description_, by default None

        Returns
        -------
        tuple[dict[str, Any], GroupedList]
            best viable association and associated modality order
        """
        # values to groupby indices with
        indices_to_groupby = [
            self._combination_formatter(combination) for combination in combinations
        ]

        # grouping tab by its indices
        grouped_xaggs = [
            self._grouper(xagg, index_to_groupby)
            for index_to_groupby in tqdm(
                indices_to_groupby, disable=not self.verbose, desc="Grouping modalities   "
            )
        ]

        # computing associations for each tabs
        n_obs = xagg.apply(sum).sum()  # number of observations for xtabs
        associations_xagg = [
            self._association_measure(grouped_xagg, n_obs=n_obs)
            for grouped_xagg in tqdm(
                grouped_xaggs, disable=not self.verbose, desc="Computing associations"
            )
        ]

        # adding corresponding combination to the association
        for combination, index_to_groupby, association, grouped_xagg in zip(
            combinations, indices_to_groupby, associations_xagg, grouped_xaggs
        ):
            association.update(
                {
                    "combination": combination,
                    "index_to_groupby": index_to_groupby,
                    "xagg": grouped_xagg,
                }
            )

        # sorting associations according to specified metric
        associations_xagg = (
            DataFrame(associations_xagg)
            .sort_values(self.sort_by, ascending=False)
            .to_dict(orient="records")
        )

        # testing viability of combination
        best_association = self._test_viability(feature, order, associations_xagg, xagg_dev, dropna)

        # applying best_combination to order and xtab
        if best_association is not None:
            order = order_apply_combination(order, best_association["combination"])

        return best_association, order

    def _test_viability(
        self,
        feature: str,
        order: GroupedList,
        associations_xagg: list[dict[str, Any]],
        xagg_dev: Union[Series, DataFrame],
        dropna: bool,
    ) -> dict[str, Any]:
        """Tests the viability of all possible combinations onto xagg_dev

        Parameters
        ----------
        feature : str
            _description_
        order : GroupedList
            order of the feature
        associations_xagg : list[dict[str, Any]]
            _description_
        xagg_dev : Union[Series, DataFrame]
            _description_

        Returns
        -------
        dict[str, Any]
            _description_
        """
        # testing viability of all combinations
        best_association, train_viable, dev_viable = (None,) * 3
        test_results: dict[str, bool] = {}
        for n_combination, association in tqdm(
            enumerate(associations_xagg),
            total=len(associations_xagg),
            disable=not self.verbose,
            desc="Testing robustness    ",
        ):
            # computing target rate and frequency per value
            train_rates = self._printer(association["xagg"])

            # viability on train sample:
            # - target rates are distinct for consecutive modalities
            distinct_rates_train = not any(
                isclose(train_rates["target_rate"][1:], train_rates["target_rate"].shift(1)[1:])
            )
            # - minimum frequency is reached for all modalities
            min_freq_train = all(train_rates["frequency"] >= self.min_freq_mod)

            # checking for viability on train
            train_viable = min_freq_train and distinct_rates_train
            test_results.update(
                {
                    "train_viable": train_viable,
                    "min_freq_train": min_freq_train,
                    "distinct_rates_train": distinct_rates_train,
                }
            )
            if train_viable:
                # case 0: no test sample provided -> not testing for robustness
                if xagg_dev is None:
                    best_association = association  # found best viable combination

                # case 1: test sample provided -> testing robustness
                else:
                    # grouping the dev sample per modality
                    grouped_xagg_dev = self._grouper(xagg_dev, association["index_to_groupby"])

                    # computing target rate and frequency per modality
                    dev_rates = self._printer(grouped_xagg_dev)

                    # viability on dev sample:
                    # - grouped values have the same ranks in train/test
                    ranks_train_dev = all(
                        train_rates.sort_values("target_rate").index
                        == dev_rates.sort_values("target_rate").index
                    )
                    # - minimum frequency is reached for all modalities
                    min_freq_dev = all(dev_rates["frequency"] >= self.min_freq_mod)
                    # - target rates are distinct for all modalities
                    distinct_rates_dev = not any(
                        isclose(dev_rates["target_rate"][1:], dev_rates["target_rate"].shift(1)[1:])
                    )

                    # checking for viability on dev
                    dev_viable = ranks_train_dev and min_freq_dev and distinct_rates_dev
                    test_results.update(
                        {
                            "dev_viable": dev_viable,
                            "min_freq_dev": min_freq_dev,
                            "distinct_rates_dev": distinct_rates_dev,
                        }
                    )
                    if dev_viable:
                        best_association = association  # found best viable combination

            # historizing combinations and tests
            self._historize_viability_test(
                feature=feature,
                association=association,
                order=order,
                n_combination=n_combination,
                associations_xagg=associations_xagg,
                dropna=dropna,
                **test_results,
            )

            # best combination found: breaking the loop on combinations
            if best_association is not None:
                break

        return best_association

    def _historize_viability_test(
        self,
        feature: str,
        association: dict[Any],
        order: GroupedList,
        n_combination: int = None,
        associations_xagg: list[dict[str, Any]] = None,
        train_viable: bool = None,
        dev_viable: bool = None,
        dropna: bool = False,
        **viability_msg_params,
    ) -> None:
        """historizes the viability tests results for specified feature

        Parameters
        ----------
        feature : str
            feature for which to historize the combination
        viability : bool
            result of viability test
        order : GroupedList
            order of the modalities and there respective groups
        association : dict[Any]
            index_to_groupby and self.sort_by values
        viability_msg_params : dict
            kwargs to determine the viability message
        """
        # Messages associated to each failed viability test
        messages = []
        if not viability_msg_params.get("ranks_train_dev", True):
            messages += ["X_dev: inversion of target rates per modality"]
        if not viability_msg_params.get("min_freq_dev", True):
            messages += [
                f"X_dev: non-representative modality (min_freq_mod={self.min_freq_mod:2.2%})"
            ]
        if not viability_msg_params.get("distinct_rates_dev", True):
            messages += ["X_dev: non-distinct target rates per consecutive modalities"]
        if not viability_msg_params.get("min_freq_train", True):
            messages += [f"X: non-representative modality (min_freq_mod={self.min_freq_mod:2.2%})"]
        if not viability_msg_params.get("distinct_rates_train", True):
            messages += ["X: non-distinct target rates per consecutive modalities"]

        # viability has been checked on train
        viability = None
        if train_viable is not None:
            # viability on train
            viability = train_viable
            if train_viable:
                # viability has been checked on dev
                if dev_viable is not None:
                    # viability on dev
                    viability = dev_viable
                    if dev_viable:
                        messages = ["Combination robust between X and X_dev"]
                else:  # no x_dev provided
                    messages = ["Combination viable on X"]
        else:
            messages = ["Raw X distribution"]

        # viability not checked for following less associated combinations
        associations_not_checked = []
        if viability:
            associations_not_checked = associations_xagg[n_combination + 1 :]

        # storing combination and adding not tested combinations to the set to be historized
        associations_to_historize = [association] + associations_not_checked
        messages_to_historize = [messages] + [["Not checked"]] * len(associations_not_checked)
        viability_to_historize = [viability] + [None] * len(associations_not_checked)

        # historizing all combinations
        for asso, msg, viab in zip(
            associations_to_historize, messages_to_historize, viability_to_historize
        ):
            # Formats a combination for historization
            combi = asso["index_to_groupby"]
            formatted_combi = [
                [
                    value
                    for modality in combi.keys()
                    for group_modality in order.get(modality, modality)
                    for value in self.values_orders[feature].get(group_modality, group_modality)
                    if combi[modality] == final_group
                ]
                for final_group in unique(list(combi.values()))
            ]

            # historizing test results
            self._history[feature] += [
                {
                    "combination": formatted_combi,
                    self.sort_by: asso[self.sort_by],
                    "viability": viab,
                    "viability_message": msg,
                    "grouping_nan": dropna,
                }
            ]

    def history(self, feature: str = None) -> DataFrame:
        """Historic of tested combinations and there association with the target.

        By default:

            * Modality ``str_default="__OTHER__"`` is generated for features that contain non-representative modalities.
            * Modality ``str_nan="__NAN__"`` is generated for features that contain ``numpy.nan``.
            * Whatever the value of ``dropna``, the association is computed for non-missing values.

        Parameters
        ----------
        feature : str, optional
            Specify for which feature to return the history, by default ``None``

        Returns
        -------
        DataFrame
            Historic of features' tested combinations.
        """
        # getting feature's history
        if feature is not None:
            assert (
                feature in self._history.keys()
            ), f"Carving of feature {feature} was not requested."
            histo = self._history[feature]

        # getting all features' history
        else:
            histo = []
            for feature in self._history.keys():
                feature_histories = self._history[feature]
                for feature_history in feature_histories:
                    feature_history.update({"feature": feature})
                histo += feature_histories

        # formatting combinations
        # history["combination"] = history["combination"].apply(format_for_history)

        return DataFrame(histo)

    def _index_mapper(
        self, feature: str, labels_orders: dict[str, GroupedList], xtab: DataFrame = None
    ) -> DataFrame:
        """Prints a binary xtab's statistics

        Parameters
        ----------
        order_get : Callable
            Ordering of modalities used to map indices
        xtab : Dataframe
            A crosstab, by default None

        Returns
        -------
        DataFrame
            Target rate and frequency per modality
        """
        # checking for an xtab
        mapped_xtab = None
        if xtab is not None:
            # copying initial xtab
            mapped_xtab = xtab.copy()

            # for qualitative features -> mapping with values_orders.content
            if feature in self.qualitative_features:
                mapped_index = [
                    self.values_orders[feature].get(idx, idx) for idx in mapped_xtab.index
                ]
                # removing str_default
                mapped_index = [
                    [str(idx) for idx in mapped_idx if idx != self.str_default]
                    for mapped_idx in mapped_index
                ]
                mapped_index = [
                    mapped_idx[-1] + " to " + mapped_idx[0]
                    if len(mapped_idx) > 2
                    else mapped_idx[0]
                    if len(mapped_idx) == 0
                    else ", ".join(mapped_idx)
                    for mapped_idx in mapped_index
                ]
            # for quantitative features -> mapping with labels_orders.keys
            else:
                mapped_index = labels_orders[feature][:]

            # modifying indices based on provided order
            mapped_xtab.index = mapped_index

        return mapped_xtab

    def _print_xagg(
        self,
        xagg: DataFrame,
        message: str,
        *,
        xagg_dev: DataFrame = None,
    ) -> None:
        """Prints crosstabs' target rates and frequencies per modality, in raw or html format

        Parameters
        ----------
        xagg : DataFrame
            Train crosstab
        xagg_dev : DataFrame
            Dev crosstab, by default None
        pretty_print : bool, optional
            Whether to output html or not, by default False
        """
        if self.verbose:  # verbose if requested
            print(f"\n - [AutoCarver] {message}")

            # getting pretty xtabs
            # TODO: remove digits from dataframes
            nice_xagg = self._printer(xagg)
            nice_xagg_dev = self._printer(xagg_dev)

            # case 0: no pretty hmtl printing
            if not self.pretty_print:
                print(nice_xagg, "\n")
                if xagg_dev is not None:
                    print("X_dev distribution\n", nice_xagg_dev, "\n")

            # case 1: pretty html printing
            else:
                # getting prettier xtabs
                nicer_xagg = prettier_xagg(nice_xagg, caption="X distribution")
                nicer_xagg_dev = prettier_xagg(nice_xagg_dev, caption="X_dev distribution")

                # merging outputs
                nicer_xaggs = nicer_xagg + "    " + nicer_xagg_dev

                # displaying html of colored DataFrame
                display_html(nicer_xaggs, raw=True)


def filter_nan(xagg: Union[Series, DataFrame], str_nan: str) -> DataFrame:
    """Filters out nans from crosstab or y values

    Parameters
    ----------
    xagg : Union[Series, DataFrame]
        _description_
    str_nan : str
        _description_

    Returns
    -------
    Union[Series, DataFrame]
        _description_
    """
    # cehcking for values in crosstab
    filtered_xagg = None
    if xagg is not None:
        # filtering out nans if requested from train crosstab
        filtered_xagg = xagg.copy()
        if str_nan in xagg.index:
            filtered_xagg = xagg.drop(str_nan, axis=0)

    return filtered_xagg


def combinations_at_index(
    start_idx: int, order: list[Any], nb_remaining_groups: int, min_group_size: int = 1
) -> tuple[list[Any], int, int]:
    """Gets all possible combinations of sizes up to the last element of a list

    Parameters
    ----------
    start_idx : int
        _description_
    order : list[Any]
        _description_
    nb_remaining_groups : int
        _description_
    min_group_size : int, optional
        _description_, by default 1

    Returns
    -------
    tuple[list[Any], int, int]
        _description_

    Yields
    ------
    Iterator[tuple[list[Any], int, int]]
        _description_
    """
    # iterating over each possible length of groups
    for size in range(min_group_size, len(order) + 1):
        next_idx = start_idx + size  # index from which to start the next group

        # checking that next index is not off the order list
        if next_idx < len(order) + 1:
            # checking that there are remaining groups or that it is the last group
            if (nb_remaining_groups > 1) | (next_idx == len(order)):
                combination = list(order[start_idx:next_idx])
                yield (combination, next_idx, nb_remaining_groups - 1)


def consecutive_combinations(
    raw_order: list[Any],
    max_group_size: int,
    min_group_size: int = 1,
    nb_remaining_group: int = None,
    current_combination: list[Any] = None,
    next_index: int = None,
    all_combinations: list[list[Any]] = None,
) -> list[list[Any]]:
    """Computes all possible combinations of values of order up to max_group_size.

    Parameters
    ----------
    raw_order : list[Any]
        _description_
    max_group_size : int
        _description_
    min_group_size : int, optional
        _description_, by default 1
    nb_remaining_group : int, optional
        _description_, by default None
    current_combination : list[Any], optional
        _description_, by default None
    next_index : int, optional
        _description_, by default None
    all_combinations : list[list[Any]], optional
        _description_, by default None

    Returns
    -------
    list[list[Any]]
        _description_
    """
    # initiating recursive attributes
    if current_combination is None:
        current_combination = []
    if next_index is None:
        next_index = 0
    if nb_remaining_group is None:
        nb_remaining_group = max_group_size
    if all_combinations is None:
        all_combinations = []

    # getting combinations for next index
    next_combinations = [
        elt
        for elt in combinations_at_index(next_index, raw_order, nb_remaining_group, min_group_size)
    ]

    # stop case: no next_combinations possible -> adding to all_combinations
    if len(next_combinations) == 0 and min_group_size < len(current_combination) <= max_group_size:
        # saving up combination
        all_combinations += [current_combination]

        # resetting remaining number of groups
        nb_remaining_group = max_group_size

    # otherwise: adding all next_combinations to the current_combination
    for combination, next_index, current_nb_remaining_group in next_combinations:
        # going a rank further in the raw_xtab
        consecutive_combinations(
            raw_order,
            max_group_size,
            min_group_size=min_group_size,
            nb_remaining_group=current_nb_remaining_group,
            current_combination=current_combination + [combination],
            next_index=next_index,
            all_combinations=all_combinations,
        )

    return all_combinations


def nan_combinations(
    raw_order: GroupedList,
    str_nan: str,
    max_n_mod: int,
    min_group_size: int,
) -> list[list[str]]:
    """All consecutive combinatios of non-nans with added nan to each possible group and a last
      group only with nan if the max_n_mod is not reached by the combination

    Parameters
    ----------
    raw_order : GroupedList
        _description_
    str_nan : str
        _description_
    max_n_mod : int
        _description_
    min_group_size : int
        _description_

    Returns
    -------
    list[list[str]]
        _description_
    """
    # all possible consecutive combinations
    combinations = consecutive_combinations(raw_order, max_n_mod, min_group_size=1)
    # iterating over each combination
    nan_combinations = []
    for combination in combinations:
        # adding nan to each group of the combination
        nan_combination = []
        for n in range(len(combination)):
            # copying input combination
            new_combination = combination[:]
            # adding nan to the nth group
            new_combination[n] = new_combination[n] + [str_nan]
            # storing updated combination with attributed group to nan
            nan_combination += [new_combination]

        # if max_n_mod is not reached adding a combination with nans alone
        if len(combination) < max_n_mod:
            # copying input combination
            new_combination = combination[:]
            # adding a group for nans only
            nan_combination += [new_combination + [[str_nan]]]

        nan_combinations += nan_combination

    return nan_combinations


def order_apply_combination(order: GroupedList, combination: list[list[Any]]) -> GroupedList:
    """Converts a list of combination to a GroupedList

    Parameters
    ----------
    order : GroupedList
        _description_
    combination : list[list[Any]]
        _description_

    Returns
    -------
    GroupedList
        _description_
    """
    order_copy = GroupedList(order)
    for combi in combination:
        order_copy.group_list(combi, combi[0])

    return order_copy


def xagg_apply_order(xagg: DataFrame, order: GroupedList) -> DataFrame:
    """Applies an order (combination) to a crosstab

    Parameters
    ----------
    xagg : DataFrame
        Crosstab
    order : GroupedList
        Combination of index to apply to the crosstab

    Returns
    -------
    dict[str, Any]
        Orderd crosstab.
    """
    # checking for input values
    combi_xagg = None
    if xagg is not None:
        # grouping modalities in the crosstab
        groups = list(map(order.get_group, xagg.index))
        combi_xagg = xagg.groupby(groups, dropna=False, sort=False).sum()

    return combi_xagg


def load_carver(auto_carver_json: dict) -> BaseDiscretizer:
    """Allows one to load an AutoCarver saved as a .json file.

    The AutoCarver has to be saved with ``json.dump(f, AutoCarver.to_json())``, otherwise there
    can be no guarantee for it to be restored.

    Parameters
    ----------
    auto_carver_json : str
        Loaded .json file using ``json.load(f)``.

    Returns
    -------
    BaseDiscretizer
        A fitted AutoCarver.
    """
    # deserializing values_orders
    values_orders = json_deserialize_values_orders(auto_carver_json["values_orders"])

    # updating auto_carver attributes
    auto_carver_json.update({"values_orders": values_orders})

    # initiating BaseDiscretizer
    auto_carver = BaseDiscretizer(**auto_carver_json)
    auto_carver.fit()

    return auto_carver


def prettier_xagg(
    nice_xagg: DataFrame = None,
    caption: str = None,
) -> str:
    """Converts a crosstab to the HTML format, adding nice colors

    Parameters
    ----------
    nice_xagg : DataFrame, optional
        Target rate and frequency per modality, by default None
    caption : str, optional
        Title of the HTML table, by default None

    Returns
    -------
    str
        HTML format of the crosstab
    """
    """Pretty display of frequency and target rate per modality on the same line."""
    # checking for a provided xtab
    nicer_xagg = ""
    if nice_xagg is not None:
        # checking for non unique indices
        if any(nice_xagg.index.duplicated()):
            nice_xagg.reset_index(inplace=True)

        # adding coolwarm color gradient
        nicer_xagg = nice_xagg.style.background_gradient(cmap="coolwarm")

        # printing inline notebook
        nicer_xagg = nicer_xagg.set_table_attributes("style='display:inline'")

        # adding custom caption/title
        if caption is not None:
            nicer_xagg = nicer_xagg.set_caption(caption)

        # converting to html
        nicer_xagg = nicer_xagg._repr_html_()

    return nicer_xagg



"""Tools to select the best Quantitative and Qualitative features for a Regression task."""

from typing import Callable

from ..discretizers import extend_docstring
from .base_selector import BaseSelector
from .filters.qualitative_filters import tschuprowt_filter
from .filters.quantitative_filters import spearman_filter
from .measures.base_measures import reverse_xy
from .measures.quantitative_measures import distance_measure, kruskal_measure


class RegressionSelector(BaseSelector):
    """A pipeline of measures to perform a feature pre-selection that maximizes association
    with a quantitative target.

    Get your best features with ``RegressionSelector.select()``!
    """

    @extend_docstring(BaseSelector.__init__)
    def __init__(
        self,
        n_best: int,
        qualitative_features: list[str] = None,
        quantitative_features: list[str] = None,
        *,
        quantitative_measures: list[Callable] = None,
        qualitative_measures: list[Callable] = None,
        quantitative_filters: list[Callable] = None,
        qualitative_filters: list[Callable] = None,
        colsample: float = 1.0,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        # default measures
        if quantitative_measures is None:
            quantitative_measures = [distance_measure]
        if qualitative_measures is None:
            qualitative_measures = [reverse_xy(kruskal_measure)]
        measures = {"float": quantitative_measures, "str": qualitative_measures}

        # default filters
        if quantitative_filters is None:
            quantitative_filters = [spearman_filter]
        if qualitative_filters is None:
            qualitative_filters = [tschuprowt_filter]
        filters = {"float": quantitative_filters, "str": qualitative_filters}

        # initiating BaseSelector with the corresponding list of measures
        super().__init__(
            n_best=n_best,
            quantitative_features=quantitative_features,
            qualitative_features=qualitative_features,
            measures=measures,
            filters=filters,
            colsample=colsample,
            verbose=verbose,
            **kwargs,
        )



""" Loads Discretization tools."""

from .discretizers import Discretizer, QualitativeDiscretizer, QuantitativeDiscretizer
from .utils.base_discretizers import BaseDiscretizer, extend_docstring
from .utils.grouped_list import GroupedList
from .utils.qualitative_discretizers import (
    CategoricalDiscretizer,
    ChainedDiscretizer,
    OrdinalDiscretizer,
)
from .utils.quantitative_discretizers import ContinuousDiscretizer
from .utils.type_discretizers import StringDiscretizer



""" Loads Carving base tools."""

from .base_carver import BaseCarver, load_carver
from .binary_carver import AutoCarver, BinaryCarver
from .continuous_carver import ContinuousCarver
from .multiclass_carver import MulticlassCarver



""" Filters based on association measures between Quantitative features.
"""

from typing import Any

from numpy import inf, nan, ones, triu
from pandas import DataFrame, notna
from statsmodels.stats.outliers_influence import variance_inflation_factor


def spearman_filter(
    X: DataFrame, ranks: DataFrame, thresh_corr: float = 1, **params
) -> dict[str, Any]:
    """Computes maximum Spearman's rho between X and X (quantitative).
    Features too correlated to a feature more associated with the target
    are excluded (according to provided ``ranks``).

    Parameters
    ----------
    X : DataFrame
        Contains columns named after ``ranks``'s index (feature names)
    ranks : DataFrame
        Ranked features as index of the association table
    thresh_corr : float, optional
        Maximum Spearman's rho bewteen features, by default ``1``

    Returns
    -------
    dict[str, Any]
        Maximum Spearman's rho with a better features
    """

    # applying quantitative filter with spearman correlation
    return quantitative_filter(X, ranks, "spearman", thresh_corr, **params)


def pearson_filter(
    X: DataFrame, ranks: DataFrame, thresh_corr: float = 1, **params
) -> dict[str, Any]:
    """Computes maximum Pearson's r between X and X (quantitative).
    Features too correlated to a feature more associated with the target
    are excluded (according to provided ``ranks``).

    Parameters
    ----------
    X : DataFrame
        Contains columns named after ``ranks``'s index (feature names)
    ranks : DataFrame
        Ranked features as index of the association table
    thresh_corr : float, optional
        Maximum Pearson's r bewteen features, by default ``1``

    Returns
    -------
    dict[str, Any]
        Maximum Pearson's r with a better feature
    """

    # applying quantitative filter with spearman correlation
    return quantitative_filter(X, ranks, "pearson", thresh_corr, **params)


# TODO
# def vif_filter(X: DataFrame, ranks: DataFrame, **params) -> dict[str, Any]:
#     """Computes Variance Inflation Factor (multicolinearity)

#     Parameters
#     ----------
#     thresh_vif, float: default inf
#       Maximum VIF between features
#     """

#     # accessing the prefered order
#     prefered_order = ranks.index

#     # initiating list association per feature
#     associations = []

#     # list of dropped features
#     dropped = []

#     # iterating over each column
#     for i, feature in enumerate(prefered_order):
#         # identifying remaining more associated features
#         better_features = [f for f in prefered_order[: i + 1] if f not in dropped]

#         X_vif = X[better_features]  # keeping only better features
#         X_vif = X_vif.dropna(axis=0)  # dropping NaNs for OLS

#         # computation of VIF
#         vif = nan
#         if len(better_features) > 1 and len(X_vif) > 0:
#             vif = variance_inflation_factor(X_vif.values, len(better_features) - 1)

#         # dropping the feature if it was too correlated to a better feature
#         if vif > params.get("thresh_vif", inf) and notna(vif):
#             dropped += [feature]

#         # kept feature: updating associations with this feature
#         else:
#             associations += [{"feature": feature, "vif_filter": vif}]

#     # formatting ouput to DataFrame
#     associations = DataFrame(associations).set_index("feature")

#     # applying filter on association
#     associations = ranks.join(associations, how="right")

#     return associations


def quantitative_filter(
    X: DataFrame, ranks: DataFrame, corr_measure: str, thresh_corr: float = 1, **params
) -> dict[str, Any]:
    """Computes max association between X and X (quantitative) excluding features
    that are correlated to a feature more associated with the target
    (defined by the ranks).

    Parameters
    ----------
    thresh_corr, float: default 1.
        Maximum association between features
    """

    # accessing the prefered order
    prefered_order = ranks.index

    # computing correlation between features
    X_corr = X[prefered_order].corr(corr_measure).abs()
    X_corr = X_corr.where(triu(ones(X_corr.shape), k=1).astype(bool))

    # initiating list of maximum association per feature
    associations = []

    # iterating over each feature by target association order
    for feature in prefered_order:
        # correlation with features more associated to the target
        corr_with_better_features = X_corr.loc[:feature, feature]

        # maximum correlation with a better feature
        corr_with, worst_corr = corr_with_better_features.agg(["idxmax", "max"])

        # dropping the feature if it was too correlated to a better feature
        if worst_corr > thresh_corr:
            X_corr = X_corr.drop(feature, axis=0).drop(feature, axis=1)

        # kept feature: updating associations with this feature
        else:
            associations += [
                {
                    "feature": feature,
                    f"{corr_measure}_filter": worst_corr,
                    f"{corr_measure}_with": corr_with,
                }
            ]

    # checking for some selected features
    if len(associations) > 0:
        # formatting ouput to DataFrame
        associations = DataFrame(associations).set_index("feature")

        # applying filter on association
        associations = ranks.join(associations, how="right")

    return associations



""" Filters based on association measure between features and a binary target.
"""

from typing import Any

from pandas import DataFrame


def thresh_filter(X: DataFrame, ranks: DataFrame, **kwargs) -> dict[str, Any]:
    """Filters out missing association measure (did not pass a threshold)"""

    # drops rows with nans
    associations = ranks.dropna(axis=0)

    return associations



""" Loads association measures."""

from .base_measures import dtype_measure, make_measure, mode_measure, nans_measure
from .qualitative_measures import chi2_measure, cramerv_measure, tschuprowt_measure
from .quantitative_measures import (
    R_measure,
    distance_measure,
    iqr_measure,
    kruskal_measure,
    zscore_measure,
)



""" Base measures that defines Quantitative and Qualitative features.
"""
from typing import Any, Callable

from pandas import Series


def reverse_xy(measure: Callable):
    """Reverses places of x and y in measure"""

    def reversed_measure(
        y: Series,
        x: Series,
        **kwargs,
    ) -> tuple[bool, dict[str, Any]]:
        """Reversed version of the measure"""
        return measure(y, x, **kwargs)

    return reversed_measure


def make_measure(
    measure: Callable,
    active: bool,
    association: dict[str, Any],
    x: Series,
    y: Series,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Wrapper to make measures from base metrics

    Parameters
    ----------
    measure : Callable
        _description_
    active : bool
        _description_
    association : dict[str, Any]
        _description_
    x : Series
        Feature to measure
    y : Series
        Binary target feature

    Returns
    -------
    tuple[bool, dict[str, Any]]
        _description_
    """
    # check that previous steps where passed for computational optimization
    if active:
        # use the measure
        active, measurement = measure(x, y, **kwargs)

        # update association table
        association.update(measurement)

    return active, association


def nans_measure(
    x: Series,
    y: Series = None,
    thresh_nan: float = 0.999,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Measure of the percentage of NaNs

    Parameters
    ----------
    x : Series
        Feature to measure
    y : Series, optional
        Binary target feature, by default ``None``
    thresh_nan : float, optional
        Maximum percentage of NaNs in a feature, by default ``0.999``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether or not there are to many NaNs and the percentage of NaNs
    """
    nans = x.isnull()  # ckecking for nans
    pct_nan = nans.mean()  # Computing percentage of nans

    # updating association
    measurement = {"pct_nan": pct_nan}

    # Excluding feature that have to many NaNs
    active = pct_nan < thresh_nan
    if not active:
        print(
            f"Feature {x.name} will be discarded (more than {thresh_nan:2.2%} of nans). Otherwise, set a greater value for thresh_nan."
        )

    return active, measurement


def dtype_measure(
    x: Series,
    y: Series = None,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Feature's dtype

    Parameters
    ----------
    x : Series
        Feature to measure
    y : Series, optional
        Binary target feature, by default ``None``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        True and the feature's dtype
    """
    # getting dtype
    measurement = {"dtype": x.dtype}

    return True, measurement


def mode_measure(
    x: Series,
    y: Series = None,
    thresh_mode: float = 0.999,
    **kwargs,
) -> tuple[bool, dict[str, Any]]:
    """Measure of the percentage of the Mode

    Parameters
    ----------
    x : Series
        Feature to measure
    y : Series, optional
        Binary target feature, by default ``None``
    thresh_mode : float, optional
        Maximum percentage of a feature's mode, by default ``0.999``

    Returns
    -------
    tuple[bool, dict[str, Any]]
        Whether or not the mode is overrepresented and the percentage of mode
    """
    mode = x.mode(dropna=True).values[0]  # computing mode
    pct_mode = (x == mode).mean()  # Computing percentage of the mode

    # updating association
    measurement = {"pct_mode": pct_mode, "mode": mode}

    # Excluding feature with too frequent modes
    active = pct_mode < thresh_mode
    if not active:
        print(
            f"Feature {x.name} will be discarded (more than {thresh_mode:2.2%} of its mode). Otherwise, set a greater value for thresh_mode."
        )

    return active, measurement



""" Loads Carving tools."""

from .carvers.base_carver import BaseCarver, load_carver
from .carvers.binary_carver import AutoCarver, BinaryCarver
from .carvers.continuous_carver import ContinuousCarver
from .carvers.multiclass_carver import MulticlassCarver



""" Loads association filters."""

from .base_filters import thresh_filter
from .qualitative_filters import cramerv_filter, tschuprowt_filter
from .quantitative_filters import pearson_filter, spearman_filter



""" Loads Discretization base tools."""

from .base_discretizers import BaseDiscretizer, extend_docstring
from .grouped_list import GroupedList
from .qualitative_discretizers import CategoricalDiscretizer, ChainedDiscretizer, OrdinalDiscretizer
from .quantitative_discretizers import ContinuousDiscretizer
from .type_discretizers import StringDiscretizer



""" Loads feature selection tools."""

from .base_selector import *
from .classification_selector import ClassificationSelector
from .filters import *
from .measures import *
from .regression_selector import RegressionSelector



""" Loads association measures."""

from ...selectors.measures.base_measures import (
    dtype_measure,
    make_measure,
    mode_measure,
    nans_measure,
)
from ...selectors.measures.qualitative_measures import (
    chi2_measure,
    cramerv_measure,
    tschuprowt_measure,
)
from ...selectors.measures.quantitative_measures import (
    R_measure,
    iqr_measure,
    kruskal_measure,
    zscore_measure,
)
